#
#   Copyright (c) 2010-2014, MIT Probabilistic Computing Project
#
#   Lead Developers: Jay Baxter and Dan Lovell
#   Authors: Jay Baxter, Dan Lovell, Baxter Eaves, Vikash Mansinghka
#   Research Leads: Vikash Mansinghka, Patrick Shafto
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import time
import inspect
import pickle
import os
import pytest
from pyparsing import *

from bayesdb.bql_grammar import *
from bayesdb.engine import Engine
from bayesdb.parser import Parser
import bayesdb.functions as functions
import numpy
engine = Engine('local')
parser = Parser()

test_M_c = {'idx_to_name': {'1': 'b', '0': 'a', '3': 'd', '2': 'c'},
            'column_metadata': [
                    {'code_to_value': {'a': 0, '1': 1, '2': 2, '4': 3, '6': 4},
                     'value_to_code': {0: 'a', 1: '1', 2: '2', 3: '4', 4: '6'},
                     'modeltype': 'symmetric_dirichlet_discrete'},
                    {'code_to_value': {}, 'value_to_code': {},
                     'modeltype': 'normal_inverse_gamma'},
                    {'code_to_value': {'we': 0, 'e': 1, 'w': 2, 'sd': 3},
                     'value_to_code': {0: 'we', 1: 'e', 2: 'w', 3: 'sd'},
                     'modeltype': 'symmetric_dirichlet_discrete'},
                    {'code_to_value': {'3': 1, '2': 2, '5': 0, '4': 3},
                     'value_to_code': {0: '5', 1: '3', 2: '2', 3: '4'},
                     'modeltype': 'symmetric_dirichlet_discrete'}],
            'name_to_idx': {'a': 0, 'c': 2, 'b': 1, 'd': 3}}

test_M_c_full = {'idx_to_name': {'1': 'b', '0': 'a', '3': 'd', '2': 'c', '4': 'key'},
            'column_metadata': [
                    {'code_to_value': {'a': 0, '1': 1, '2': 2, '4': 3, '6': 4},
                     'value_to_code': {0: 'a', 1: '1', 2: '2', 3: '4', 4: '6'},
                     'modeltype': 'symmetric_dirichlet_discrete'},
                    {'code_to_value': {}, 'value_to_code': {},
                     'modeltype': 'normal_inverse_gamma'},
                    {'code_to_value': {'we': 0, 'e': 1, 'w': 2, 'sd': 3},
                     'value_to_code': {0: 'we', 1: 'e', 2: 'w', 3: 'sd'},
                     'modeltype': 'symmetric_dirichlet_discrete'},
                    {'code_to_value': {'3': 1, '2': 2, '5': 0, '4': 3},
                     'value_to_code': {0: '5', 1: '3', 2: '2', 3: '4'},
                     'modeltype': 'symmetric_dirichlet_discrete'},
                    {'code_to_value': {'0': 0, '1': 1, '2': 2, '3': 3, '4': 4, '5': 5, '6': 6, '7': 7, '8': 8},
                     'value_to_code': {0: '0', 1: '1', 2: '2', 3: '3', 4: '4', 5: '5', 6: '6', 7: '7', 8: '8'},
                     'modeltype': 'ignore'}],
            'name_to_idx': {'a': 0, 'c': 2, 'b': 1, 'd': 3, 'key': 4}}


test_T = [[1.0, 1.0, 0.0, numpy.nan],
          [2.0, 2.0, 0.0, 2.0],
          [0.0, 3.0, 0.0, 3.0],
          [3.0, 3.0, 2.0, numpy.nan],
          [3.0, 4.0, 2.0, 0.0],
          [4.0, 5.0, 1.0, numpy.nan],
          [numpy.nan, 6.0, 2.0, 1.0],
          [numpy.nan, 7.0, 3.0, 1.0],
          [numpy.nan, 7.0, 3.0, 1.0]]

test_T_full = [[1.0, 1.0, 0.0, numpy.nan, 0],
          [2.0, 2.0, 0.0, 2.0, 1],
          [0.0, 3.0, 0.0, 3.0, 2],
          [3.0, 3.0, 2.0, numpy.nan, 3],
          [3.0, 4.0, 2.0, 0.0, 4],
          [4.0, 5.0, 1.0, numpy.nan, 5],
          [numpy.nan, 6.0, 2.0, 1.0, 6],
          [numpy.nan, 7.0, 3.0, 1.0, 7],
          [numpy.nan, 7.0, 3.0, 1.0, 8]]

def test_keyword_plurality_ambiguity_pyparsing():
    model = model_keyword.parseString("model",parseAll=True)
    models = model_keyword.parseString("models",parseAll=True)
    assert model[0] == 'model'
    assert models[0] == 'model'
    iteration = iteration_keyword.parseString("iteration",parseAll=True)
    iterations = iteration_keyword.parseString("iterations",parseAll=True)
    assert iteration[0] == 'iteration'
    assert iterations[0] == 'iteration'
    sample = sample_keyword.parseString("sample",parseAll=True)
    samples = sample_keyword.parseString("samples",parseAll=True)
    assert sample[0] == 'sample'
    assert samples[0] == 'sample'
    column = column_keyword.parseString('column',parseAll=True)
    columns = column_keyword.parseString('columns',parseAll=True)
    assert column[0] == 'column'
    assert columns[0] == 'column'
    list_ = list_keyword.parseString('list',parseAll=True)
    lists = list_keyword.parseString('lists',parseAll=True)
    assert list_[0] == 'list'
    assert lists[0] == 'list'
    btable = btable_keyword.parseString('btable',parseAll=True)
    btables = btable_keyword.parseString('btables',parseAll=True)
    assert btable[0] == 'btable'
    assert btables[0] == 'btable'
    minute = minute_keyword.parseString('minute',parseAll=True)
    minutes = minute_keyword.parseString('minutes',parseAll=True)
    assert minute[0] == 'minute'
    assert minute[0] == 'minute'

def test_composite_keywords_pyparsing():
    execute_file = execute_file_keyword.parseString('eXecute file',parseAll=True)
    assert execute_file[0] == 'execute_file'
    create_btable = create_btable_keyword.parseString('cReate btable',parseAll=True)
    assert create_btable[0] == 'create_btable'
    update_schema_for = update_schema_for_keyword.parseString('update Schema for',parseAll=True)
    assert update_schema_for[0] == 'update_schema'
    models_for = models_for_keyword.parseString('Models for',parseAll=True)
    assert models_for[0] == 'model for'
    model_index = model_index_keyword.parseString('model Index',parseAll=True)
    assert model_index[0] == 'model index'
    save_model = save_model_keyword.parseString("save modeL",parseAll=True)
    assert save_model[0] == 'save_models'
    load_model = load_model_keyword.parseString("load Models",parseAll=True)
    assert load_model[0] == 'load_models'
    save_to = save_to_keyword.parseString('save To',parseAll=True)
    assert save_to[0] == 'save to'
    list_btables = list_btables_keyword.parseString('list bTables',parseAll=True)
    assert list_btables[0] == 'list_btables'
    show_schema_for = show_schema_for_keyword.parseString('show Schema for',parseAll=True)
    assert show_schema_for[0] == 'show_schema'
    show_models_for = show_models_for_keyword.parseString("show modeLs for",parseAll=True)
    assert show_models_for[0] == 'show_models'
    show_diagnostics_for = show_diagnostics_for_keyword.parseString("show diaGnostics for",parseAll=True)
    assert show_diagnostics_for[0] == 'show_diagnostics'
    estimate_pairwise = estimate_pairwise_keyword.parseString("estimate Pairwise",parseAll=True)
    assert estimate_pairwise[0] == 'estimate_pairwise'
    with_confidence = with_confidence_keyword.parseString('with  confIdence',parseAll=True)
    assert with_confidence[0] == 'with confidence'
    dependence_probability = dependence_probability_keyword.parseString('dependence probability',parseAll=True)
    assert dependence_probability[0] == 'dependence probability'
    mutual_information = mutual_information_keyword.parseString('mutual inFormation',parseAll=True)
    assert mutual_information[0] == 'mutual information'
    estimate_columns_from = estimate_columns_from_keyword.parseString("estimate columns froM",parseAll=True)
    assert estimate_columns_from[0] == 'estimate_columns'
    column_lists = column_lists_keyword.parseString('column Lists',parseAll=True)
    assert column_lists[0] == 'column list'
    with_respect_to = with_respect_to_keyword.parseString("with Respect to",parseAll=True)
    assert with_respect_to[0] == 'with respect to'
    probability_of = probability_of_keyword.parseString('probability of',parseAll=True)
    assert probability_of[0] == 'probability'
    predictive_probability_of = predictive_probability_of_keyword.parseString('predictive Probability  of',parseAll=True)
    assert predictive_probability_of[0] == 'predictive probability'
    save_clusters_with_threshold = save_clusters_with_threshold_keyword.parseString(
        'save clusters with threshold',parseAll=True)
    assert save_clusters_with_threshold[0] == 'save clusters with threshold'
    estimate_pairwise_row = estimate_pairwise_row_keyword.parseString("estimate Pairwise row",parseAll=True)
    assert estimate_pairwise_row[0] == 'estimate_pairwise_row'

def test_valid_values_names_pyparsing():
    valid_values=[
        '4',
        '42.04',
        '.4',
        '4.',
        "'\sjekja8391(*^@(%()!@#$%^&*()_+=-~'",
        "a0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ!\"#$%&'()*+-./:<=>?@[\]^_`{|}~",
        'b0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ!"#$%&\'()*+-./:<=>?@[\]^_`{|}~',
        '"c0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ!\\"#$%&\'()*+-./:<=>?@[\]^_`{|}~"',
        "'d0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ!\"#$%&\\'()*+-./:<=>?@[\]^_`{|}~'",
        "'numbers 0'",
        "'k skj s'",
        ]
    valid_values_results=[
        '4',
        '42.04',
        '.4',
        '4.',
        '\sjekja8391(*^@(%()!@#$%^&*()_+=-~',
        "a0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ!\"#$%&'()*+-./:<=>?@[\]^_`{|}~",
        'b0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ!"#$%&\'()*+-./:<=>?@[\]^_`{|}~',
        "c0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ!\"#$%&\'()*+-./:<=>?@[\]^_`{|}~",
        "d0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ!\"#$%&\'()*+-./:<=>?@[\]^_`{|}~",
        'numbers 0',
        'k skj s',
        ]

    for i in range(len(valid_values)):
        assert value.parseString(valid_values[i],parseAll=True)[0] == valid_values_results[i]

    valid_column_identifiers = [
        "a",
        "a1",
        "a_1",
        "a_a",
        "a_",
        "aa"
        ]
    valid_column_identifiers_results = [
        "a",
        "a1",
        "a_1",
        "a_a",
        "a_",
        "aa"
        ]
    for i in range(len(valid_column_identifiers)):
        assert value.parseString(valid_column_identifiers[i],parseAll=True)[0] == valid_column_identifiers_results[i]
    assert float_number.parseString('1',parseAll=True)[0] == '1'
    assert int_number.parseString('1',parseAll=True)[0] == '1'
    assert float_number.parseString('1.')[0] == '1'
    assert float_number.parseString('.1',parseAll=True)[0] == '.1'
    assert float_number.parseString('0.1',parseAll=True)[0] == '0.1'
    assert float_number.parseString('11',parseAll=True)[0] == '11'
    assert int_number.parseString('11',parseAll=True)[0] == '11'
    assert float_number.parseString('11.01',parseAll=True)[0] == '11.01'
    assert filename.parseString("~/filename.csv",parseAll=True)[0] == "~/filename.csv"
    assert filename.parseString("!\"/#$%&'()*+-.:<=>?@[\]^_`{|}~",parseAll=True)[0] == "!\"/#$%&'()*+-.:<=>?@[\]^_`{|}~"
    assert filename.parseString("'/filename with space.csv'",parseAll=True)[0] == "/filename with space.csv"

def test_simple_functions():
    assert list_btables_function.parseString("LIST BTABLES",parseAll=True).statement_id == 'list_btables'
    assert list_btables_function.parseString("LIST BTABLE",parseAll=True).statement_id == 'list_btables'
    assert show_for_btable_statement.parseString("SHOW SCHEMA FOR table_1",parseAll=True).statement_id == 'show_schema'
    assert show_for_btable_statement.parseString("SHOW SCHEMA FOR table_1",parseAll=True).btable == 'table_1'
    assert show_for_btable_statement.parseString("SHOW MODELS FOR table_1",parseAll=True).statement_id == 'show_models'
    assert show_for_btable_statement.parseString("SHOW MODEL FOR table_1",parseAll=True).btable == 'table_1'
    assert show_for_btable_statement.parseString("SHOW DIAGNOSTICS FOR table_1",parseAll=True).statement_id == 'show_diagnostics'
    assert show_for_btable_statement.parseString("SHOW DIAGNOSTICS FOR table_1",parseAll=True).btable == 'table_1'

    assert show_for_btable_statement.parseString("SHOW COLUMN LISTS FOR table_1",parseAll=True).btable == 'table_1'
    assert show_for_btable_statement.parseString("SHOW COLUMNS LIST FOR table_1",parseAll=True).statement_id == 'show_column_lists'
    assert show_columns_function.parseString("SHOW COLUMNS asdf FOR table_1",parseAll=True).column_list == 'asdf'
    assert drop_column_list_function.parseString("DROP COLUMN LIST group1 FROM table_1", parseAll=True).list_name == 'group1'
    assert drop_column_list_function.parseString("DROP COLUMN LIST group1 FROM table_1", parseAll=True).statement_id == 'drop_column_list'
    assert drop_column_list_function.parseString("DROP COLUMN LIST group1 FROM table_1", parseAll=True).btable == 'table_1'
    assert show_for_btable_statement.parseString("SHOW ROW LISTS FOR table_1",parseAll=True).statement_id == 'show_row_lists'
    assert show_for_btable_statement.parseString("SHOW ROW list FOR table_1",parseAll=True).btable == 'table_1'
    assert drop_row_list_function.parseString("DROP ROW LIST group1 FROM table_1", parseAll=True).list_name == 'group1'
    assert drop_row_list_function.parseString("DROP ROW LIST group1 FROM table_1", parseAll=True).statement_id == 'drop_row_list'
    assert drop_row_list_function.parseString("DROP ROW LIST group1 FROM table_1", parseAll=True).btable == 'table_1'
    assert load_model_function.parseString("LOAD MODELS ~/filename.csv INTO table_1",parseAll=True).statement_id == 'load_models'
    assert load_model_function.parseString("LOAD MODEL ~/filename.csv INTO table_1",parseAll=True).statement_id == 'load_models'
    assert load_model_function.parseString("LOAD MODELS ~/filename.csv INTO table_1",parseAll=True).filename == '~/filename.csv'
    assert load_model_function.parseString("LOAD MODELS '~/filena me.csv' INTO table_1",parseAll=True).filename == '~/filena me.csv'
    assert load_model_function.parseString("LOAD MODELS ~/filename.csv INTO table_1",parseAll=True).btable == 'table_1'
    assert save_model_from_function.parseString("SAVE MODEL FROM table_1 to filename.pkl.gz",parseAll=True).btable == 'table_1'
    assert save_model_from_function.parseString("SAVE MODEL FROM table_1 to filename.pkl.gz",parseAll=True).statement_id == 'save_models'
    assert save_model_from_function.parseString("SAVE MODEL FROM table_1 to filename.pkl.gz",parseAll=True).filename == 'filename.pkl.gz'
    assert drop_btable_function.parseString("DROP BTABLE table_1",parseAll=True).statement_id == 'drop_btable'
    assert drop_btable_function.parseString("DROP BTABLES table_1",parseAll=True).statement_id == 'drop_btable'
    assert drop_btable_function.parseString("DROP BTABLE table_1",parseAll=True).btable == 'table_1'
    drop_model_1 = drop_model_function.parseString("DROP MODEL 1 FROM table_1",parseAll=True)
    drop_model_2 = drop_model_function.parseString("DROP MODELS 1-5 FROM table_1",parseAll=True)
    drop_model_3 = drop_model_function.parseString("DROP MODELS 1,2,6-9 FROM table_1",parseAll=True)
    drop_model_4 = drop_model_function.parseString("DROP MODELS 1-5,1-5 FROM table_1",parseAll=True)
    assert drop_model_1.statement_id == 'drop_models'
    assert drop_model_1.btable == 'table_1'
    assert drop_model_1.index_clause.asList() == [1]
    assert drop_model_2.index_clause.asList() == [1,2,3,4,5]
    assert drop_model_3.index_clause.asList() == [1,2,6,7,8,9]
    assert drop_model_4.index_clause.asList() == [1,2,3,4,5]
    assert help_function.parseString("HELp",parseAll=True).statement_id == 'help'

def test_update_schema_pyparsing():
    update_schema_1 = update_schema_for_function.parseString("UPDATE SCHEMA FOR test_btable SET col_1 = Categorical,col_2=numerical , col_3  =  ignore",parseAll=True)
    assert update_schema_1.statement_id == 'update_schema'
    assert update_schema_1.btable == 'test_btable'
    assert update_schema_1.type_clause[0][0] == 'col_1'
    assert update_schema_1.type_clause[0][1] == 'categorical'
    assert update_schema_1.type_clause[1][0] == 'col_2'
    assert update_schema_1.type_clause[1][1] == 'numerical'
    assert update_schema_1.type_clause[2][0] == 'col_3'
    assert update_schema_1.type_clause[2][1] == 'ignore'
    update_schema_2 = update_schema_for_function.parseString("UPDATE SCHEMA FOR test_btable SET col_1 = key",parseAll=True)
    assert update_schema_2.type_clause[0][0] == 'col_1'
    assert update_schema_2.type_clause[0][1] == 'key'
    update_schema_3 = update_schema_for_function.parseString("UPDATE SCHEMA FOR test_btable SET col_1 = categorical(15)", parseAll=True)
    assert update_schema_3.type_clause[0][0] == 'col_1'
    assert update_schema_3.type_clause[0][1] == 'categorical'
    assert update_schema_3.type_clause[0].parameters.cardinality == '15'
    update_schema_4 = update_schema_for_function.parseString("UPDATE SCHEMA FOR test_btable SET col_2 = cyclic(0, 10)", parseAll=True)
    assert update_schema_4.type_clause[0][0] == 'col_2'
    assert update_schema_4.type_clause[0][1] == 'cyclic'
    assert update_schema_4.type_clause[0].parameters.min == '0'
    assert update_schema_4.type_clause[0].parameters.max == '10'

def test_update_codebook_pyparsing():
    bql_string = "UPDATE CODEBOOK FOR test_btable FROM new_codebook.csv"
    update_codebook = update_codebook_for_function.parseString(bql_string)

    assert update_codebook.statement_id == 'update_codebook'
    assert update_codebook.btable == 'test_btable'
    assert update_codebook.filename == 'new_codebook.csv'

def test_describe_pyparsing():
    bql_string = "DESCRIBE col_1 FOR test_btable"
    describe_0 = describe_function.parseString(bql_string)
    bql_string = "DESCRIBE col_1, col_2 FOR test_btable"
    describe_1 = describe_function.parseString(bql_string)

    assert describe_0.statement_id == 'describe'
    assert describe_0.btable == 'test_btable'
    assert 'col_1' in describe_0.columnset.asList()

    assert describe_1.statement_id == 'describe'
    assert describe_1.btable == 'test_btable'
    assert 'col_1' in describe_1.columnset.asList()
    assert 'col_2' in describe_1.columnset.asList()

def test_update_descriptions_pyparsing():
    bql_string = 'UPDATE DESCRIPTION FOR test_btable SET col_1="Hamish the cat"'
    updated_description_0 = update_descriptions_for_function.parseString(bql_string)

    assert updated_description_0.statement_id == "update_descriptions"
    assert updated_description_0.btable == "test_btable"
    assert updated_description_0.label_clause[0][0] == "col_1"
    assert updated_description_0.label_clause[0][1] == "Hamish the cat"

    bql_string = 'UPDATE DESCRIPTIONS FOR test_btable SET col_1="Hamish the cat", col_2="trevor"'
    updated_description_1 = update_descriptions_for_function.parseString(bql_string)

    assert updated_description_1.statement_id == "update_descriptions"
    assert updated_description_1.btable == "test_btable"
    assert updated_description_1.label_clause[0][0] == "col_1"
    assert updated_description_1.label_clause[0][1] == "Hamish the cat"
    assert updated_description_1.label_clause[1][0] == "col_2"
    assert updated_description_1.label_clause[1][1] == "trevor"

def test_update_short_names_pyparsing():
    bql_string = 'UPDATE SHORT NAME FOR test_btable SET col_1="Hamish"'
    updated_short_names_0 = update_short_names_for_function.parseString(bql_string)

    assert updated_short_names_0.statement_id == "update_short_names"
    assert updated_short_names_0.btable == "test_btable"
    assert updated_short_names_0.label_clause[0][0] == "col_1"
    assert updated_short_names_0.label_clause[0][1] == "Hamish"

    bql_string = 'UPDATE SHORT NAMES FOR test_btable SET col_1="Hamish", col_2="trevor"'
    updated_short_names_1 = update_short_names_for_function.parseString(bql_string)

    assert updated_short_names_1.statement_id == "update_short_names"
    assert updated_short_names_1.btable == "test_btable"
    assert updated_short_names_1.label_clause[0][0] == "col_1"
    assert updated_short_names_1.label_clause[0][1] == "Hamish"
    assert updated_short_names_1.label_clause[1][0] == "col_2"
    assert updated_short_names_1.label_clause[1][1] == "trevor"

def test_create_btable_pyparsing():
    create_btable_1 = create_btable_function.parseString("CREATE BTABLE test.btable FROM '~/filenam e.csv'", parseAll=True)
    create_btable_2 = create_btable_function.parseString("CREATE BTABLE test_btable FROM ~/filename.csv", parseAll=True)
    assert create_btable_1.statement_id == 'create_btable'
    assert create_btable_1.btable == 'test.btable'
    assert create_btable_1.filename == '~/filenam e.csv'
    assert create_btable_2.btable == 'test_btable'
    assert create_btable_2.filename == '~/filename.csv'

def test_execute_file_pyparsing():
    execute_file_1 = execute_file_function.parseString("EXECUTE FILE '/filenam e.bql'",parseAll=True)
    execute_file_2 = execute_file_function.parseString("EXECUTE FILE /filename.bql",parseAll=True)
    assert execute_file_1.filename == "/filenam e.bql"
    assert execute_file_2.filename == "/filename.bql"

def test_initialize_pyparsing():
    initialize_1 = initialize_function.parseString("INITIALIZE 3 MODELS FOR test_table",parseAll=True)
    assert initialize_1.statement_id == 'initialize_models'
    assert initialize_1.num_models == '3'
    assert initialize_1.btable == 'test_table'
    initialize_2 = initialize_function.parseString("INITIALIZE 3 MODEL FOR test_table",parseAll=True)
    assert initialize_2.statement_id == 'initialize_models'
    assert initialize_2.num_models == '3'
    assert initialize_2.btable == 'test_table'
    initialize_3 = initialize_function.parseString("INITIALIZE MODELS FOR test_table",parseAll=True)
    assert initialize_3.statement_id == 'initialize_models'
    assert initialize_3.num_models == ''
    assert initialize_3.btable == 'test_table'

def test_analyze_pyparsing():
    analyze_1 = analyze_function.parseString("ANALYZE table_1 FOR 10 ITERATIONS",parseAll=True)
    analyze_2 = analyze_function.parseString("ANALYZE table_1 FOR 1 ITERATION",parseAll=True)
    analyze_3 = analyze_function.parseString("ANALYZE table_1 FOR 10 MINUTES",parseAll=True)
    analyze_4 = analyze_function.parseString("ANALYZE table_1 FOR 1 MINUTE",parseAll=True)
    analyze_5 = analyze_function.parseString("ANALYZE table_1 MODEL 1 FOR 10 MINUTES",parseAll=True)
    analyze_6 = analyze_function.parseString("ANALYZE table_1 MODELS 1-3 FOR 1 ITERATION",parseAll=True)
    analyze_7 = analyze_function.parseString("ANALYZE table_1 MODELS 1,2,3 FOR 10 MINUTES",parseAll=True)
    analyze_8 = analyze_function.parseString("ANALYZE table_1 MODELS 1, 3-5 FOR 1 ITERATION",parseAll=True)
    analyze_9 = analyze_function.parseString("ANALYZE table_1 MODELS 1-3, 5 FOR 10 MINUTES",parseAll=True)
    analyze_10 = analyze_function.parseString("ANALYZE table_1 MODELS 1-3, 5-7, 9, 10 FOR 1 ITERATION",parseAll=True)
    analyze_11 = analyze_function.parseString("ANALYZE table_1 MODELS 1, 1, 2, 2 FOR 10 MINUTES",parseAll=True)
    analyze_12 = analyze_function.parseString("ANALYZE table_1 MODELS 1-5, 1-5, 5 FOR 1 ITERATION",parseAll=True)
    assert analyze_1.statement_id == 'analyze'
    assert analyze_1.btable == 'table_1'
    assert analyze_1.index_lust == ''
    assert analyze_1.index_clause == ''
    assert analyze_1.num_iterations == '10'
    assert analyze_1.num_minutes == ''
    assert analyze_2.num_iterations == '1'
    assert analyze_2.num_minutes == ''
    assert analyze_3.num_iterations == ''
    assert analyze_3.num_minutes == '10'
    assert analyze_4.num_iterations == ''
    assert analyze_4.num_minutes == '1'
    assert analyze_5.index_clause.asList() == [1]
    assert analyze_6.index_clause.asList() == [1,2,3]
    assert analyze_7.index_clause.asList() == [1,2,3]
    assert analyze_8.index_clause.asList() == [1,3,4,5]
    assert analyze_9.index_clause.asList() == [1,2,3,5]
    assert analyze_10.index_clause.asList() == [1,2,3,5,6,7,9,10]
    assert analyze_11.index_clause.asList() == [1,2]
    assert analyze_12.index_clause.asList() == [1,2,3,4,5]

def test_subclauses_pyparsing():
    assert save_to_clause.parseString("save to filename.csv").filename == 'filename.csv'

def test_row_clause_pyparsing():
    row_1 = row_clause.parseString('1', parseAll=True)
    row_2 = row_clause.parseString("column = 1", parseAll=True)
    row_3 = row_clause.parseString("column = 'value'", parseAll=True)
    row_4 = row_clause.parseString("column = value", parseAll=True)
    assert row_1.row_id == '1'
    assert row_1.column == ''
    assert row_2.row_id == ''
    assert row_2.column == 'column'
    assert row_2.column_value == '1'
    assert row_3.column_value == 'value'
    assert row_4.column_value == 'value'

def test_row_functions_pyparsing():
    similarity_1 = similarity_to_function.parseString("SIMILARITY TO 1",
                                                      parseAll=True)
    similarity_2 = similarity_to_function.parseString("SIMILARITY TO col_2 = 1",
                                                      parseAll=True)
    similarity_3 = similarity_to_function.parseString("SIMILARITY TO col_2 = 'a'",
                                                      parseAll=True)
    similarity_4 = similarity_to_function.parseString("SIMILARITY TO col_2 = a",
                                                      parseAll=True)
    similarity_5 = similarity_to_function.parseString("SIMILARITY TO 1 WITH RESPECT TO col_1",
                                                      parseAll=True)
    similarity_6 = similarity_to_function.parseString("SIMILARITY TO col_2 = 1 WITH RESPECT TO col_1,col_2",
                                                      parseAll=True)
    similarity_7 = similarity_to_function.parseString("SIMILARITY TO col_2 = 'a' WITH RESPECT TO col_1 , col_3",
                                                      parseAll=True)
    similarity_8 = similarity_to_function.parseString("SIMILARITY TO col_2 = a WITH RESPECT TO col_1",
                                                      parseAll=True)
    assert similarity_1.function.function_id == 'similarity'
    assert similarity_1.function.row_id == '1'
    assert similarity_2.function.column == 'col_2'
    assert similarity_2.function.column_value == '1'
    assert similarity_3.function.column == 'col_2'
    assert similarity_3.function.column_value == 'a'
    assert similarity_4.function.column == 'col_2'
    assert similarity_4.function.column_value == 'a'
    assert similarity_4.function.with_respect_to == ''
    assert not similarity_5.function.with_respect_to == ''
    assert similarity_5.function.column_list.asList() == ['col_1']
    assert similarity_6.function.column_list.asList() == ['col_1', 'col_2']
    assert similarity_7.function.column_list.asList() == ['col_1', 'col_3']
    assert similarity_8.function.column_list.asList() == ['col_1']
    assert typicality_function.parseString('Typicality',parseAll=True).function.function_id == 'typicality'

def test_column_functions_pyparsing():
    dependence_1 = dependence_probability_function.parseString('DEPENDENCE PROBABILITY WITH column_1',
                                                                    parseAll=True)
    dependence_2 = dependence_probability_function.parseString('DEPENDENCE PROBABILITY OF column_2 WITH column_1',
                                                                    parseAll=True)
    assert dependence_1.function.function_id == 'dependence probability'
    assert dependence_2.function.function_id == 'dependence probability'
    assert dependence_1.function.with_column == 'column_1'
    assert dependence_2.function.with_column == 'column_1'
    assert dependence_2.function.of_column == 'column_2'
    mutual_1 = mutual_information_function.parseString('MUTUAL INFORMATION WITH column_1',
                                                                    parseAll=True)
    mutual_2 = mutual_information_function.parseString('MUTUAL INFORMATION OF column_2 WITH column_1',
                                                                    parseAll=True)
    assert mutual_1.function.function_id == 'mutual information'
    assert mutual_2.function.function_id == 'mutual information'
    assert mutual_1.function.with_column == 'column_1'
    assert mutual_2.function.with_column == 'column_1'
    assert mutual_2.function.of_column == 'column_2'
    correlation_1 = correlation_function.parseString('CORRELATION WITH column_1',
                                                                    parseAll=True)
    correlation_2 = correlation_function.parseString('CORRELATION OF column_2 WITH column_1',
                                                                    parseAll=True)
    assert correlation_1.function.function_id == 'correlation'
    assert correlation_2.function.function_id == 'correlation'
    assert correlation_1.function.with_column == 'column_1'
    assert correlation_2.function.with_column == 'column_1'
    assert correlation_2.function.of_column == 'column_2'


def test_probability_of_function_pyparsing():
    probability_of_1 = probability_of_function.parseString("PROBABILITY OF col_1 = 1",parseAll=True)
    probability_of_2 = probability_of_function.parseString("PROBABILITY OF col_1 = 'value'",parseAll=True)
    probability_of_3 = probability_of_function.parseString("PROBABILITY OF col_1 = value",parseAll=True)
    assert probability_of_1.function.function_id == 'probability'
    assert probability_of_1.function.column == 'col_1'
    assert probability_of_1.function.value == '1'
    assert probability_of_2.function.value == 'value'
    assert probability_of_3.function.value == 'value'

def test_predictive_probability_of_pyparsing():
    assert predictive_probability_of_function.parseString("PREDICTIVE PROBABILITY OF column_1",
                                                          parseAll=True).function.function_id == 'predictive probability'
    assert predictive_probability_of_function.parseString("PREDICTIVE PROBABILITY OF column_1",
                                                          parseAll=True).function.column == 'column_1'

def test_typicality_of_pyparsing():
    assert typicality_function.parseString("TYPICALITY OF column_1",
                                                          parseAll=True).function.function_id == 'typicality'
    assert typicality_function.parseString("TYPICALITY OF column_1",
                                                          parseAll=True).function.column == 'column_1'

def test_order_by_clause_pyparsing():
    order_by_1 = order_by_clause.parseString("ORDER BY column_1"
                                             ,parseAll=True)
    order_by_2 = order_by_clause.parseString("ORDER BY column_1,column_2 , column_3"
                                             ,parseAll=True)
    assert order_by_1.order_by[0].function.column == 'column_1'
    assert order_by_2.order_by[1].function.column =='column_2'
    order_by_3 = order_by_clause.parseString("ORDER BY TYPICALITY",
                                             parseAll=True)
    assert order_by_3.order_by[0].function.function_id == 'typicality'
    order_by_4 = order_by_clause.parseString("ORDER BY TYPICALITY, column_1",
                                             parseAll=True)
    assert order_by_4.order_by[0].function.function_id == 'typicality'
    assert order_by_4.order_by[1].function.column == 'column_1'
    order_by_5 = order_by_clause.parseString("ORDER BY column_1, TYPICALITY",
                                             parseAll=True)
    assert order_by_5.order_by[0].function.column == 'column_1'
    assert order_by_5.order_by[1].function.function_id == 'typicality'
    order_by_6 = order_by_clause.parseString("ORDER BY PREDICTIVE PROBABILITY OF column_1",
                                             parseAll=True)
    assert order_by_6.order_by[0].function.function_id == 'predictive probability'
    assert order_by_6.order_by[0].function.column == 'column_1'

    order_by_7 = order_by_clause.parseString("ORDER BY PREDICTIVE PROBABILITY OF column_1, column_1",
                                             parseAll=True)
    assert order_by_7.order_by[1].function.column == 'column_1'
    assert order_by_7.order_by[0].function.function_id == 'predictive probability'
    assert order_by_7.order_by[0].function.column == 'column_1'

    order_by_8 = order_by_clause.parseString("ORDER BY column_1, TYPICALITY, PREDICTIVE PROBABILITY OF column_1, column_2, SIMILARITY TO 2, SIMILARITY TO column_1 = 1 WITH RESPECT TO column_4",
                                             parseAll=True)
    assert order_by_8.order_by[0].function.column == 'column_1'
    assert order_by_8.order_by[1].function.function_id == 'typicality'
    assert order_by_8.order_by[2].function.function_id == 'predictive probability'
    assert order_by_8.order_by[2].function.column == 'column_1'
    assert order_by_8.order_by[3].function.column == 'column_2'
    assert order_by_8.order_by[4].function.function_id == 'similarity'
    assert order_by_8.order_by[4].function.row_id == '2'
    assert order_by_8.order_by[5].function.function_id == 'similarity'
    assert order_by_8.order_by[5].function.column == 'column_1'
    assert order_by_8.order_by[5].function.column_value == '1'
    assert order_by_8.order_by[5].function.with_respect_to[1][0] == 'column_4' #todo names instead of indexes

    order_by_9 = order_by_clause.parseString("ORDER BY column_1 asc"
                                             ,parseAll=True)
    order_by_10 = order_by_clause.parseString("ORDER BY column_1 asc,column_2 desc , column_3"
                                             ,parseAll=True)
    assert order_by_9.order_by[0].function.column =='column_1'
    assert order_by_10.order_by[1].function.column =='column_2'
    assert order_by_9.order_by[0].asc_desc =='asc'
    assert order_by_10.order_by[1].asc_desc =='desc'
    order_by_11 = order_by_clause.parseString("ORDER BY TYPICALITY asc",
                                             parseAll=True)
    assert order_by_11.order_by[0].asc_desc =='asc'

def test_whereclause_pyparsing():
    # WHERE <column> <operation> <value>
    whereclause_1 = "WHERE column_1 = 1"
    parsed_1 = where_clause.parseString(whereclause_1,parseAll=True)
    assert parsed_1.where_keyword == 'where'
    assert parsed_1.where_conditions[0].function.column == 'column_1'
    assert parsed_1.where_conditions[0].operation == '='
    assert parsed_1.where_conditions[0].value == '1'
    whereclause_2 = "WHERE column_1 <= 1"
    parsed_2 = where_clause.parseString(whereclause_2,parseAll=True)
    assert parsed_2.where_conditions[0].function.column == 'column_1'
    assert parsed_2.where_conditions[0].operation == '<='
    assert parsed_2.where_conditions[0].value == '1'
    whereclause_3 = "WHERE column_1 > 1.0"
    parsed_3 = where_clause.parseString(whereclause_3,parseAll=True)
    assert parsed_3.where_conditions[0].operation == '>'
    assert parsed_3.where_conditions[0].value == '1.0'
    whereclause_4 = "WHERE column_1 = a"
    parsed_4 = where_clause.parseString(whereclause_4,parseAll=True)
    assert parsed_4.where_conditions[0].operation == '='
    assert parsed_4.where_conditions[0].value == 'a'
    whereclause_5 = "WHERE column_1 = 'a'"
    parsed_5 = where_clause.parseString(whereclause_5,parseAll=True)
    assert parsed_5.where_conditions[0].value == 'a'
    whereclause_6 = "WHERE column_1 = 'two words'"
    parsed_6 = where_clause.parseString(whereclause_6,parseAll=True)
    assert parsed_6.where_conditions[0].value == 'two words'
    # Functions
    whereclause_7 = "WHERE TYPICALITY > .8"
    parsed_7 = where_clause.parseString(whereclause_7,parseAll=True)
    assert parsed_7.where_conditions[0].function.function_id == 'typicality'
    assert parsed_7.where_conditions[0].operation == '>'
    assert parsed_7.where_conditions[0].value == '.8'
    whereclause_8 = "WHERE PREDICTIVE PROBABILITY OF column_1 > .1"
    parsed_8 = where_clause.parseString(whereclause_8,parseAll=True)
    assert parsed_8.where_conditions[0].function.function_id == 'predictive probability'
    assert parsed_8.where_conditions[0].function.column == 'column_1'
    assert parsed_8.where_conditions[0].operation == '>'
    assert parsed_8.where_conditions[0].value == '.1'
    whereclause_9 = "WHERE SIMILARITY TO 2 > .1"
    parsed_9 = where_clause.parseString(whereclause_9,parseAll=True)
    assert parsed_9.where_conditions[0].function.function_id == 'similarity'
    assert parsed_9.where_conditions[0].function.row_id == '2'
    assert parsed_9.where_conditions[0].operation == '>'
    assert parsed_9.where_conditions[0].value == '.1'
    whereclause_10 = "WHERE SIMILARITY TO 2 WITH RESPECT TO column_1 > .4"
    parsed_10 = where_clause.parseString(whereclause_10,parseAll=True)
    assert parsed_10.where_conditions[0].function.function_id == 'similarity'
    assert parsed_10.where_conditions[0].function.row_id == '2'
    assert parsed_10.where_conditions[0].function.with_respect_to.column_list[0] == 'column_1'
    assert parsed_10.where_conditions[0].operation == '>'
    assert parsed_10.where_conditions[0].value == '.4'
    whereclause_11 = "WHERE SIMILARITY TO column_1 = 1 = .5"
    parsed_11 = where_clause.parseString(whereclause_11,parseAll=True)
    assert parsed_11.where_conditions[0].function.function_id == 'similarity'
    assert parsed_11.where_conditions[0].function.column == 'column_1'
    assert parsed_11.where_conditions[0].function.column_value == '1'
    assert parsed_11.where_conditions[0].operation == '='
    assert parsed_11.where_conditions[0].value == '.5'
    whereclause_12 = "WHERE SIMILARITY TO column_1 = 'a' WITH RESPECT TO column_2 > .5"
    parsed_12 = where_clause.parseString(whereclause_12,parseAll=True)
    assert parsed_12.where_conditions[0].function.function_id == 'similarity'
    assert parsed_12.where_conditions[0].function.column == 'column_1'
    assert parsed_12.where_conditions[0].function.column_value == 'a'
    assert parsed_12.where_conditions[0].operation == '>'
    assert parsed_12.where_conditions[0].value == '.5'
    assert parsed_12.where_conditions[0].function.with_respect_to.column_list[0] == 'column_2'
    whereclause_13 = "WHERE SIMILARITY TO column_1 = 1.2 WITH RESPECT TO column_2 > .5"
    parsed_13 = where_clause.parseString(whereclause_13,parseAll=True)
    assert parsed_13.where_conditions[0].function.function_id == 'similarity'
    assert parsed_13.where_conditions[0].function.column == 'column_1'
    assert parsed_13.where_conditions[0].function.column_value == '1.2'
    assert parsed_13.where_conditions[0].operation == '>'
    assert parsed_13.where_conditions[0].value == '.5'
    assert parsed_13.where_conditions[0].function.with_respect_to.column_list[0] == 'column_2'
    whereclause_14 = "WHERE SIMILARITY TO column_1 = a WITH RESPECT TO column_2 > .5"
    parsed_14 = where_clause.parseString(whereclause_14,parseAll=True)
    assert parsed_14.where_conditions[0].function.function_id == 'similarity'
    assert parsed_14.where_conditions[0].function.column == 'column_1'
    assert parsed_14.where_conditions[0].function.column_value == 'a'
    assert parsed_14.where_conditions[0].operation == '>'
    assert parsed_14.where_conditions[0].value == '.5'
    assert parsed_14.where_conditions[0].function.with_respect_to.column_list[0] == 'column_2'
    # With Confidence
    whereclause_15 = "WHERE TYPICALITY > .8 CONF .5"
    parsed_15 = where_clause.parseString(whereclause_15,parseAll=True)
    assert parsed_15.where_conditions[0].conf == '.5'
    whereclause_16 = "WHERE PREDICTIVE PROBABILITY OF column_1 > .1 CONF .5"
    parsed_16 = where_clause.parseString(whereclause_16,parseAll=True)
    assert parsed_16.where_conditions[0].conf == '.5'
    whereclause_17 = "WHERE SIMILARITY TO 2 > .1 CONF .5"
    parsed_17 = where_clause.parseString(whereclause_17,parseAll=True)
    assert parsed_17.where_conditions[0].conf == '.5'
    whereclause_18 = "WHERE SIMILARITY TO 2 WITH RESPECT TO column_1 > .4 CONF .5"
    parsed_18 = where_clause.parseString(whereclause_18,parseAll=True)
    assert parsed_18.where_conditions[0].conf == '.5'
    whereclause_19 = "WHERE SIMILARITY TO column_1 = 1 = .5 CONF .5"
    parsed_19 = where_clause.parseString(whereclause_19,parseAll=True)
    assert parsed_19.where_conditions[0].conf == '.5'
    whereclause_20 = "WHERE SIMILARITY TO column_1 = 'a' WITH RESPECT TO column_2 > .5 CONF .5"
    parsed_20 = where_clause.parseString(whereclause_20,parseAll=True)
    assert parsed_20.where_conditions[0].conf == '.5'
    whereclause_21 = "WHERE SIMILARITY TO column_1 = 1.2 WITH RESPECT TO column_2 > .5 CONF .5"
    parsed_21 = where_clause.parseString(whereclause_21,parseAll=True)
    assert parsed_21.where_conditions[0].conf == '.5'
    whereclause_22 = "WHERE SIMILARITY TO column_1 = a WITH RESPECT TO column_2 > .5 CONF .5"
    parsed_22 = where_clause.parseString(whereclause_22,parseAll=True)
    assert parsed_22.where_conditions[0].conf == '.5'
    # AND
    whereclause_23 = "WHERE column_1 = 'a' AND column_2 >= 3"
    parsed_23 = where_clause.parseString(whereclause_23,parseAll=True)
    assert parsed_23.where_conditions[0].function.column == 'column_1'
    assert parsed_23.where_conditions[1].function.column == 'column_2'
    whereclause_24 = "WHERE TYPICALITY > .8 AND PREDICTIVE PROBABILITY OF column_1 > .1 AND SIMILARITY TO 2 > .1"
    parsed_24 = where_clause.parseString(whereclause_24,parseAll=True)
    assert parsed_24.where_conditions[0].function.function_id == 'typicality'
    assert parsed_24.where_conditions[1].function.function_id == 'predictive probability'
    assert parsed_24.where_conditions[2].function.function_id == 'similarity'
    whereclause_25 = "WHERE TYPICALITY > .8 CONF .4 AND PREDICTIVE PROBABILITY OF column_1 > .1 CONF .6 AND SIMILARITY TO 2 > .1 CONF .5"
    parsed_25 = where_clause.parseString(whereclause_25,parseAll=True)
    assert parsed_25.where_conditions[0].conf == '.4'
    assert parsed_25.where_conditions[1].conf == '.6'
    assert parsed_25.where_conditions[2].conf == '.5'
    whereclause_26 = "WHERE KEY IN row_list_1 AND column_1 = 'a' AND TYPICALITY > .4"
    parsed_26 = where_clause.parseString(whereclause_26,parseAll=True)
    assert parsed_26.where_conditions[0].function.function_id == 'key'
    assert parsed_26.where_conditions[0].value == 'row_list_1'
    assert parsed_26.where_conditions[1].function.column == 'column_1'
    assert parsed_26.where_conditions[2].function.function_id == 'typicality'

def test_key_in_rowlist():
    assert key_in_rowlist_clause.parseString("key in row_list_1",parseAll=True).function.function_id == "key"
    assert key_in_rowlist_clause.parseString("key in row_list_1",parseAll=True).value == "row_list_1"

def test_basic_select_pyparsing():
    select_1 = "SELECT * FROM table_1"
    select_1_parse = query.parseString(select_1,parseAll=True)
    assert select_1_parse.statement_id == 'select'
    assert select_1_parse.btable == 'table_1'
    assert select_1_parse.functions[0].column_id == '*'
    select_2 = "SELECT column_1,column_3 FROM table_1"
    select_2_parse = query.parseString(select_2,parseAll=True)
    assert select_2_parse.functions[0].column_id == 'column_1'
    assert select_2_parse.functions[1].column_id == 'column_3'
    select_3 = "PLOT SELECT column_1 FROM table_1 WHERE column_2 = 3"
    select_3_parse = query.parseString(select_3,parseAll=True)
    assert select_3_parse.plot == 'plot'
    assert select_3_parse.functions[0].column_id == 'column_1'
    assert select_3_parse.where_keyword == 'where'
    assert select_3_parse.where_conditions[0].value == '3'
    assert select_3_parse.where_conditions[0].function.column == 'column_2'
    assert select_3_parse.where_conditions[0].operation == '='
    select_4 = "SELECT col_1 FROM table_1 ORDER BY TYPICALITY LIMIT 10 SAVE TO ~/test.txt"
    select_4_parse = query.parseString(select_4,parseAll=True)
    assert select_4_parse.functions[0].column_id == 'col_1'
    assert select_4_parse.order_by[0].function.function_id == 'typicality'
    assert select_4_parse.limit == '10'
    assert select_4_parse.filename == '~/test.txt'

def test_select_functions_pyparsing():
    query_1 = "SELECT TYPICALITY FROM table_1"
    query_2 = "SELECT TYPICALITY OF column_1 FROM table_1"
    query_3 = "SELECT PREDICTIVE PROBABILITY OF column_1 FROM table_1"
    query_4 = "SELECT PROBABILITY OF column_1 = 4 FROM table_1"
    query_5 = "SELECT SIMILARITY TO 0 FROM table_1"
    query_5 = "SELECT SIMILARITY TO column_1 = 4 FROM table_1"
    query_6 = "SELECT DEPENDENCE PROBABILITY WITH column_1 FROM table_1"
    query_7 = "SELECT MUTUAL INFORMATION OF column_1 WITH column_2 FROM table_1"
    query_8 = "SELECT CORRELATION OF column_1 WITH column_2 FROM table_1"
    query_9 = "SELECT TYPICALITY, PREDICTIVE PROBABILITY OF column_1 FROM table_1"
    select_ast_1 = query.parseString(query_1,parseAll=True)
    select_ast_2 = query.parseString(query_2,parseAll=True)
    select_ast_3 = query.parseString(query_3,parseAll=True)
    select_ast_4 = query.parseString(query_4,parseAll=True)
    select_ast_5 = query.parseString(query_5,parseAll=True)
    select_ast_6 = query.parseString(query_6,parseAll=True)
    select_ast_7 = query.parseString(query_7,parseAll=True)
    select_ast_8 = query.parseString(query_8,parseAll=True)
    select_ast_9 = query.parseString(query_9,parseAll=True)
    assert select_ast_1.statement_id == 'select'
    assert select_ast_2.statement_id == 'select'
    assert select_ast_3.statement_id == 'select'
    assert select_ast_4.statement_id == 'select'
    assert select_ast_5.statement_id == 'select'
    assert select_ast_5.statement_id == 'select'
    assert select_ast_6.statement_id == 'select'
    assert select_ast_7.statement_id == 'select'
    assert select_ast_8.statement_id == 'select'
    assert select_ast_9.statement_id == 'select'
    assert select_ast_1.functions[0].function_id == 'typicality'
    assert select_ast_2.functions[0].function_id == 'typicality'
    assert select_ast_3.functions[0].function_id == 'predictive probability'
    assert select_ast_4.functions[0].function_id == 'probability'
    assert select_ast_5.functions[0].function_id == 'similarity'
    assert select_ast_5.functions[0].function_id == 'similarity'
    assert select_ast_6.functions[0].function_id == 'dependence probability'
    assert select_ast_7.functions[0].function_id == 'mutual information'
    assert select_ast_8.functions[0].function_id == 'correlation'
    assert select_ast_9.functions[0].function_id == 'typicality'
    assert select_ast_9.functions[1].function_id == 'predictive probability'

def test_infer_pyparsing():
    infer_1 = "INFER * FROM table_1"
    infer_1_parse = query.parseString(infer_1,parseAll=True)
    assert infer_1_parse.statement_id == 'infer'
    assert infer_1_parse.btable == 'table_1'
    assert infer_1_parse.functions[0].column_id == '*'
    infer_2 = "infer column_1,column_3 FROM table_1"
    infer_2_parse = query.parseString(infer_2,parseAll=True)
    assert infer_2_parse.functions[0].column_id == 'column_1'
    assert infer_2_parse.functions[1].column_id == 'column_3'
    infer_3 = "SUMMARIZE infer column_1 FROM table_1 WHERE column_2 = 3"
    infer_3_parse = query.parseString(infer_3,parseAll=True)
    assert infer_3_parse.summarize == 'summarize'
    assert infer_3_parse.functions[0].column_id == 'column_1'
    assert infer_3_parse.where_keyword == 'where'
    assert infer_3_parse.where_conditions[0].value == '3'
    assert infer_3_parse.where_conditions[0].function.column == 'column_2'
    assert infer_3_parse.where_conditions[0].operation == '='
    infer_4 = "infer col_1 FROM table_1 ORDER BY TYPICALITY LIMIT 10 SAVE TO ~/test.txt"
    infer_4_parse = query.parseString(infer_4,parseAll=True)
    assert infer_4_parse.functions[0].column_id == 'col_1'
    assert infer_4_parse.order_by[0].function.function_id == 'typicality'
    assert infer_4_parse.limit == '10'
    assert infer_4_parse.filename == '~/test.txt'
    query_1 = "INFER TYPICALITY FROM table_1 WITH CONFIDENCE .4 WITH 4 SAMPLES"
    query_2 = "INFER TYPICALITY OF column_1 FROM table_1 WITH CONFIDENCE .4 WITH 4 SAMPLES"
    query_3 = "INFER PREDICTIVE PROBABILITY OF column_1 FROM table_1 WITH CONFIDENCE .4 WITH 4 SAMPLES"
    query_4 = "INFER PROBABILITY OF column_1 = 4 FROM table_1 WITH CONFIDENCE .4 WITH 4 SAMPLES"
    query_5 = "INFER SIMILARITY TO 0 FROM table_1 WITH CONFIDENCE .4 WITH 4 SAMPLES"
    query_5 = "INFER SIMILARITY TO column_1 = 4 FROM table_1 WITH CONFIDENCE .4 WITH 4 SAMPLES"
    query_6 = "INFER DEPENDENCE PROBABILITY WITH column_1 FROM table_1 WITH CONFIDENCE .4 WITH 4 SAMPLES"
    query_7 = "INFER MUTUAL INFORMATION OF column_1 WITH column_2 FROM table_1 WITH CONFIDENCE .4 WITH 4 SAMPLES"
    query_8 = "INFER CORRELATION OF column_1 WITH column_2 FROM table_1 WITH CONFIDENCE .4 WITH 4 SAMPLES"
    query_9 = "INFER TYPICALITY, PREDICTIVE PROBABILITY OF column_1 FROM table_1 WITH CONFIDENCE .4 WITH 4 SAMPLES"
    infer_ast_1 = query.parseString(query_1,parseAll=True)
    infer_ast_2 = query.parseString(query_2,parseAll=True)
    infer_ast_3 = query.parseString(query_3,parseAll=True)
    infer_ast_4 = query.parseString(query_4,parseAll=True)
    infer_ast_5 = query.parseString(query_5,parseAll=True)
    infer_ast_6 = query.parseString(query_6,parseAll=True)
    infer_ast_7 = query.parseString(query_7,parseAll=True)
    infer_ast_8 = query.parseString(query_8,parseAll=True)
    infer_ast_9 = query.parseString(query_9,parseAll=True)
    assert infer_ast_1.statement_id == 'infer'
    assert infer_ast_2.statement_id == 'infer'
    assert infer_ast_3.statement_id == 'infer'
    assert infer_ast_4.statement_id == 'infer'
    assert infer_ast_5.statement_id == 'infer'
    assert infer_ast_5.statement_id == 'infer'
    assert infer_ast_6.statement_id == 'infer'
    assert infer_ast_7.statement_id == 'infer'
    assert infer_ast_8.statement_id == 'infer'
    assert infer_ast_9.statement_id == 'infer'
    assert infer_ast_1.functions[0].function_id == 'typicality'
    assert infer_ast_2.functions[0].function_id == 'typicality'
    assert infer_ast_3.functions[0].function_id == 'predictive probability'
    assert infer_ast_4.functions[0].function_id == 'probability'
    assert infer_ast_5.functions[0].function_id == 'similarity'
    assert infer_ast_5.functions[0].function_id == 'similarity'
    assert infer_ast_6.functions[0].function_id == 'dependence probability'
    assert infer_ast_7.functions[0].function_id == 'mutual information'
    assert infer_ast_8.functions[0].function_id == 'correlation'
    assert infer_ast_9.functions[0].function_id == 'typicality'
    assert infer_ast_9.functions[1].function_id == 'predictive probability'
    assert infer_ast_1.samples == '4'
    assert infer_ast_1.confidence == '.4'
    assert infer_ast_2.samples == '4'
    assert infer_ast_2.confidence == '.4'
    assert infer_ast_3.samples == '4'
    assert infer_ast_3.confidence == '.4'
    assert infer_ast_4.samples == '4'
    assert infer_ast_4.confidence == '.4'
    assert infer_ast_5.samples == '4'
    assert infer_ast_5.confidence == '.4'
    assert infer_ast_6.samples == '4'
    assert infer_ast_6.confidence == '.4'
    assert infer_ast_7.samples == '4'
    assert infer_ast_7.confidence == '.4'
    assert infer_ast_8.samples == '4'
    assert infer_ast_8.confidence == '.4'
    assert infer_ast_9.samples == '4'
    assert infer_ast_9.confidence == '.4'

def test_simulate_pyparsing():
    query_1 = "SIMULATE * FROM table_1 WHERE column_1 = 4 TIMES 4 SAVE TO ~/test.csv"
    simulate_ast = query.parseString(query_1,parseAll=True)
    assert simulate_ast.statement_id == 'simulate'
    assert simulate_ast.functions[0].column_id == '*'
    assert simulate_ast.where_keyword == 'where'
    assert simulate_ast.times == '4'
    assert simulate_ast.filename == '~/test.csv'
    query_2 = "SIMULATE col1,col2 FROM table_1 WHERE column_1 = 4 TIMES 4 SAVE TO ~/test.csv"
    simulate_ast = query.parseString(query_2,parseAll=True)
    assert simulate_ast.functions[0].column_id == 'col1'
    assert simulate_ast.functions[1].column_id == 'col2'

    query_3 = "SIMULATE col1, col2 FROM table_1"
    simulate_ast = query.parseString(query_3, parseAll=True)


def test_estimate_columns_from_pyparsing():
    query_1 = "ESTIMATE COLUMNS FROM table_1 WHERE col_1 = 4 ORDER BY TYPICALITY LIMIT 10 AS col_list_1"
    est_col_ast_1 = query.parseString(query_1,parseAll=True)
    assert est_col_ast_1.statement_id == 'estimate'
    assert est_col_ast_1.btable == 'table_1'
    assert est_col_ast_1.where_keyword == 'where'
    assert est_col_ast_1.where_conditions[0].function.column == 'col_1'
    assert est_col_ast_1.where_conditions[0].value == '4'
    assert est_col_ast_1.order_by[0].function.function_id == 'typicality'
    assert est_col_ast_1.limit == '10'
    assert est_col_ast_1.as_column_list == 'col_list_1'
    query_2 = "ESTIMATE COLUMNS FROM table_1"
    est_col_ast_2 = query.parseString(query_2,parseAll=True)
    assert est_col_ast_2.statement_id == 'estimate'
    assert est_col_ast_2.btable == 'table_1'

def test_estimate_pairwise_pyparsing():
    query_1 = "ESTIMATE PAIRWISE CORRELATION WITH col_1 FROM table_1"
    est_pairwise_ast_1 = query.parseString(query_1,parseAll=True)
    assert est_pairwise_ast_1.statement_id == 'estimate_pairwise'
    assert est_pairwise_ast_1.functions[0].function_id == 'correlation'
    assert est_pairwise_ast_1.functions[0].with_column == 'col_1'
    assert est_pairwise_ast_1.btable == 'table_1'
    query_2 = "ESTIMATE PAIRWISE DEPENDENCE PROBABILITY WITH col_1 FROM table_1 FOR col_1 SAVE TO file.csv SAVE CLUSTERS WITH THRESHOLD .4 AS col_list_1"
    est_pairwise_ast_2 = query.parseString(query_2,parseAll=True)
    assert est_pairwise_ast_2.statement_id == 'estimate_pairwise'
    assert est_pairwise_ast_2.functions[0].function_id == 'dependence probability'
    assert est_pairwise_ast_2.functions[0].with_column == 'col_1'
    assert est_pairwise_ast_2.btable == 'table_1'
    assert est_pairwise_ast_2.for_list.asList() == ['col_1']
    assert est_pairwise_ast_2.filename == 'file.csv'
    assert est_pairwise_ast_2.clusters_clause.threshold == '.4'
    assert est_pairwise_ast_2.clusters_clause.as_label == 'col_list_1'
    query_3 = "ESTIMATE PAIRWISE MUTUAL INFORMATION WITH col_1 FROM table_1"
    est_pairwise_ast_3 = query.parseString(query_3,parseAll=True)
    assert est_pairwise_ast_3.functions[0].function_id == 'mutual information'

def test_estimate_pairwise_row_pyparsing():
    query_1 = "ESTIMATE PAIRWISE ROW SIMILARITY FROM table_1 SAVE CLUSTERS WITH THRESHOLD .4 INTO table_2"
    est_pairwise_ast_1 = query.parseString(query_1,parseAll=True)
    assert est_pairwise_ast_1.statement_id == 'estimate_pairwise_row'
    assert est_pairwise_ast_1.functions[0].function_id == 'similarity'
    assert est_pairwise_ast_1.btable == 'table_1'
    query_2 = "ESTIMATE PAIRWISE ROW SIMILARITY FROM table_1 FOR a SAVE TO file.csv SAVE CLUSTERS WITH THRESHOLD .4 AS table_2"
    est_pairwise_ast_2 = query.parseString(query_2,parseAll=True)
    assert est_pairwise_ast_2.statement_id == 'estimate_pairwise_row'
    assert est_pairwise_ast_2.functions[0].function_id == 'similarity'
    assert est_pairwise_ast_2.btable == 'table_1'
    assert est_pairwise_ast_2.for_list.asList() == ['a']
    assert est_pairwise_ast_2.filename == 'file.csv'
    assert est_pairwise_ast_2.clusters_clause.threshold == '.4'
    assert est_pairwise_ast_2.clusters_clause.as_label == 'table_2'

def test_nested_queries_basic_pyparsing():
    query_1 = "SELECT * FROM ( SELECT col_1,col_2 FROM table_2)"
    ast = query.parseString(query_1,parseAll=True)
    assert ast.statement_id == 'select'
    assert ast.sub_query == " SELECT col_1,col_2 FROM table_2"
    ast_2 = query.parseString(ast.sub_query,parseAll=True)
    assert ast_2.statement_id == 'select'
    assert ast_2.functions[0].column_id == 'col_1'
    assert ast_2.functions[1].column_id == 'col_2'

def test_master_query_for_parse_errors():
    # This test will not test for correct information, just successfull parsing
    query_list = ["LIST BTABLES",
                  "SHOW SCHEMA FOR table_1",
                  "SHOW SCHEMA FOR table_1",
                  "SHOW DIAGNOSTICS FOR table_1",
                  "SHOW COLUMN LISTS FOR table_1",
                  "SHOW COLUMNS collist FOR table_1",
                  "LOAD MODELS ~/filename.csv INTO table_1",
                  "SAVE MODEL FROM table_1 to filename.pkl.gz",
                  "DROP BTABLE table_1",
                  "DROP MODEL 1 FROM table_1",
                  "DROP MODELS 1,2,6-9 FROM table_1",
                  "UPDATE SCHEMA FOR test_btablE SET col_1 = Categorical,col.2=numerical , col_3  =  ignore",
                  "CREATE BTABLE test.btable FROM '~/filenam e.csv'",
                  "CREATE BTABLE test_btable FROM ~/filename.csv",
                  "EXECUTE FILE '/filenam e.bql'",
                  "ANALYZE table_1 FOR 10 ITERATIONS",
                  "ANALYZE table_1 MODELS 1-3, 5-7, 9, 10 FOR 1 ITERATION",
                  "ANALYZE table_1 MODELS 1-3 FOR 1 ITERATION",
                  "SELECT TYPICALITY FROM table_1",
                  "SELECT TYPICALITY OF column_1 FROM table_1",
                  "SELECT PREDICTIVE PROBABILITY OF column_1 FROM table_1",
                  "SELECT PROBABILITY OF column_1 = 4 FROM table_1",
                  "SELECT SIMILARITY TO 0 FROM table_1",
                  "SELECT SIMILARITY TO column_1 = 4 FROM table_1",
                  "SELECT DEPENDENCE PROBABILITY WITH column_1 FROM table_1",
                  "SELECT MUTUAL INFORMATION OF column_1 WITH column_2 FROM table_1",
                  "SELECT CORRELATION OF column_1 WITH column_2 FROM table_1",
                  "SELECT TYPICALITY, PREDICTIVE PROBABILITY OF column_1 FROM table_1",
                  "SELECT SIMILARITY TO 0 WITH RESPECT TO column_1, col2 FROM table_1",
                  "SELECT SIMILARITY TO a = 1 , PROBABILITY OF a = 1 FROM table_1"]
#SELECT PREDICTIVE PROBABILITY OF a, MUTUAL INFORMATION OF a WITH b, CORRELATION OF a WITH b, DEPENDENCE PROBABILITY OF a WITH b, SIMILARITY TO 0, SIMILARITY TO a = 1, PROBABILITY OF a = 1 FROM table_1
    for query in query_list:
        query = bql_statement.parseString(query,parseAll=True)
        assert query.statement_id != ''

def test_list_btables():
    method, args, client_dict = parser.parse_single_statement(bql_statement.parseString('list btables',parseAll=True))
    assert method == 'list_btables'
    assert args == {}

def test_initialize_models():
    method, args, client_dict = parser.parse_single_statement(bql_statement.parseString('initialize 5 models for t',parseAll=True))
    assert method == 'initialize_models'
    assert args == dict(tablename='t', n_models=5, model_config=None)

def test_create_btable():
    method, args, client_dict = parser.parse_single_statement(bql_statement.parseString('create btable t from fn',parseAll=True))
    assert method == 'create_btable'
    assert args == dict(tablename='t', cctypes_full=None)
    assert client_dict == dict(csv_path=os.path.join(os.getcwd(), 'fn'))

def test_drop_btable():
    method, args, client_dict = parser.parse_single_statement(bql_statement.parseString('drop btable t',parseAll=True))
    assert method == 'drop_btable'
    assert args == dict(tablename='t')

def test_drop_models():
    method, args, client_dict = parser.parse_single_statement(bql_statement.parseString('drop models from t',parseAll=True))
    assert method == 'drop_models'
    assert args == dict(tablename='t', model_indices=None)

    method, args, client_dict = parser.parse_single_statement(bql_statement.parseString('drop models 2-6 from t'))
    assert method == 'drop_models'
    assert args == dict(tablename='t', model_indices=range(2,7))

def test_analyze():
    method, args, client_dict = parser.parse_single_statement(bql_statement.parseString('analyze t models 2-6 for 3 iterations'))
    assert method == 'analyze'
    assert args == dict(tablename='t', model_indices=range(2,7), iterations=3, seconds=None, ct_kernel=0, background=True)

    method, args, client_dict = parser.parse_single_statement(bql_statement.parseString('analyze t for 6 iterations'))
    assert method == 'analyze'
    assert args == dict(tablename='t', model_indices=None, iterations=6, seconds=None, ct_kernel=0, background=True)

    method, args, client_dict = parser.parse_single_statement(bql_statement.parseString('analyze t for 7 minutes'))
    assert method == 'analyze'

    assert args == dict(tablename='t', model_indices=None, iterations=None, seconds=7*60, ct_kernel=0, background=True)

    method, args, client_dict = parser.parse_single_statement(bql_statement.parseString('analyze t models 2-6 for 7 minutes'))
    assert method == 'analyze'
    assert args == dict(tablename='t', model_indices=range(2,7), iterations=None, seconds=7*60, ct_kernel=0, background=True)

    method, args, client_dict = parser.parse_single_statement(bql_statement.parseString('analyze t models 2-6 for 7 minutes with mh kernel'))
    assert method == 'analyze'
    assert args == dict(tablename='t', model_indices=range(2,7), iterations=None, seconds=7*60, ct_kernel=1, background=True)

def test_load_models():
    method, args, client_dict = parser.parse_single_statement(bql_statement.parseString('load models fn into t'))
    assert method == 'load_models'
    assert args == dict(tablename='t')
    assert client_dict == dict(pkl_path='fn')

def test_save_models():
    method, args, client_dict = parser.parse_single_statement(bql_statement.parseString('save models from t to fn'))
    assert method == 'save_models'
    assert args == dict(tablename='t')
    assert client_dict == dict(pkl_path='fn')

def test_parse_functions():
    query_1 = "SELECT PREDICTIVE PROBABILITY OF a, MUTUAL INFORMATION OF a WITH b, CORRELATION OF a WITH b, DEPENDENCE PROBABILITY OF a WITH b, SIMILARITY TO 0, SIMILARITY TO a = 1 , PROBABILITY OF a = 1 , probability of b = 1 , TYPICALITY of a, typicality , a , * FROM table_1"
    ast_1 = bql_statement.parseString(query_1, parseAll=True)
    function_groups = ast_1.functions

    queries, query_cols = parser.parse_functions(function_groups, M_c = test_M_c, M_c_full = test_M_c_full, T=test_T, T_full = test_T_full)
    assert queries[0] == (functions._predictive_probability, 0, False)
    assert queries[1] == (functions._mutual_information, (0,1), True)
    assert queries[2] == (functions._correlation, (0,1), True)
    assert queries[3] == (functions._dependence_probability, (0,1), True)
    assert queries[4] == (functions._similarity, (0,None), False)
    assert queries[5] == (functions._similarity, (0,None), False)
    assert queries[6] == (functions._probability, (0,'1'), True)
    assert queries[7] == (functions._probability, (1,1), True)
    assert queries[8] == (functions._col_typicality, 0, True)
    assert queries[9] == (functions._row_typicality, True, False)
    assert queries[10] == (functions._column, (0, None), False)
    assert queries[11] == (functions._column, (0, None), False)
    assert queries[12] == (functions._column, (1, None), False)

def test_select():
    ##TODO test client_dict
    tablename = 't'
    newtablename = 'newtable'
    functions = function_in_query.parseString('*',parseAll=True)
    whereclause = None
    limit = float('inf')
    order_by = False
    plot = False

    method, args, client_dict = parser.parse_single_statement(bql_statement.parseString('select * from t'))
    d = dict(tablename=tablename, functions=functions, whereclause=whereclause,
             limit=limit, order_by=order_by, plot=plot, modelids=None, summarize=False)
    assert method == 'select'
    assert args['tablename'] == d['tablename']
    assert args['functions'][0].column_id == d['functions'][0].column_id
    assert args['whereclause'] == d['whereclause']
    assert args['limit'] == d['limit']
    assert args['order_by'] == d['order_by']
    assert args['plot'] == d['plot']
    assert args['modelids'] == d['modelids']
    assert args['summarize'] == d['summarize']

    method, args, client_dict = parser.parse_single_statement(bql_statement.parseString('summarize select * from t'))

    d = dict(tablename=tablename, functions=functions, whereclause=whereclause,
             limit=limit, order_by=order_by, plot=plot, modelids=None, summarize=True)
    assert method == 'select'
    assert args['tablename'] == d['tablename']
    assert args['functions'][0].column_id == d['functions'][0].column_id
    assert args['whereclause'] == d['whereclause']
    assert args['limit'] == d['limit']
    assert args['order_by'] == d['order_by']
    assert args['plot'] == d['plot']
    assert args['modelids'] == d['modelids']
    assert args['summarize'] == d['summarize']

    method, args, client_dict = parser.parse_single_statement(bql_statement.parseString('select a, b, a_b from t'))
    d = dict(tablename=tablename, functions=None, whereclause=whereclause,
             limit=limit, order_by=order_by, plot=plot, modelids=None, summarize=False)
    assert method == 'select'
    assert args['tablename'] == d['tablename']
    assert args['functions'][0].column_id == 'a'
    assert args['functions'][1].column_id == 'b'
    assert args['functions'][2].column_id == 'a_b'
    assert args['whereclause'] == d['whereclause']
    assert args['limit'] == d['limit']
    assert args['order_by'] == d['order_by']
    assert args['plot'] == d['plot']
    assert args['modelids'] == d['modelids']
    assert args['summarize'] == d['summarize']

    method, args, client_dict = parser.parse_single_statement(bql_statement.parseString('select * from t where a=6 and b = 7'))
    d = dict(tablename=tablename, functions=functions, whereclause=whereclause,
             limit=limit, order_by=order_by, plot=plot, modelids=None, summarize=False)
    assert method == 'select'
    assert args['tablename'] == d['tablename']
    assert args['functions'][0].column_id == '*'
    assert args['whereclause'][0].function.column == 'a'
    assert args['whereclause'][0].value == '6'
    assert args['whereclause'][1].function.column == 'b'
    assert args['whereclause'][1].value == '7'
    assert args['limit'] == d['limit']
    assert args['order_by'] == d['order_by']
    assert args['plot'] == d['plot']
    assert args['modelids'] == d['modelids']
    assert args['summarize'] == d['summarize']

    limit = 10
    method, args, client_dict = parser.parse_single_statement(bql_statement.parseString('select * from t where a=6 and b = 7 limit 10'))
    d = dict(tablename=tablename, functions=functions, whereclause=whereclause,
             limit=limit, order_by=order_by, plot=plot, modelids=None, summarize=False)
    assert method == 'select'
    assert args['tablename'] == d['tablename']
    assert args['functions'][0].column_id == d['functions'][0].column_id
    assert args['whereclause'][0].function.column == 'a'
    assert args['whereclause'][0].value == '6'
    assert args['whereclause'][1].function.column == 'b'
    assert args['whereclause'][1].value == '7'
    assert args['limit'] == d['limit']
    assert args['order_by'] == d['order_by']
    assert args['plot'] == d['plot']
    assert args['modelids'] == d['modelids']
    assert args['summarize'] == d['summarize']

    method, args, client_dict = parser.parse_single_statement(bql_statement.parseString('select * from t where a=6 and b = 7 order by b limit 10'))
    order_by = [('b', True)],
    d = dict(tablename=tablename, functions=functions, whereclause=whereclause,
             limit=limit, order_by=order_by, plot=plot, modelids=None, summarize=False)
    assert method == 'select'
    assert args['tablename'] == d['tablename']
    assert args['functions'][0].column_id == d['functions'][0].column_id
    assert args['whereclause'][0].function.column == 'a'
    assert args['whereclause'][0].value == '6'
    assert args['whereclause'][1].function.column == 'b'
    assert args['whereclause'][1].value == '7'
    assert args['limit'] == d['limit']
    #assert args['order_by'] == d['order_by'] ##TODO
    assert args['plot'] == d['plot']
    assert args['modelids'] == d['modelids']
    assert args['summarize'] == d['summarize']

    method, args, client_dict = parser.parse_single_statement(bql_statement.parseString('select * from t where a=6 and b = 7 order by b limit 10 into newtable'))
    order_by = [('b', True)],
    d = dict(tablename=tablename, functions=functions, whereclause=whereclause,
             limit=limit, order_by=order_by, plot=plot, modelids=None, summarize=False, newtablename=newtablename)
    assert method == 'select'
    assert args['tablename'] == d['tablename']
    assert args['functions'][0].column_id == d['functions'][0].column_id
    assert args['whereclause'][0].function.column == 'a'
    assert args['whereclause'][0].value == '6'
    assert args['whereclause'][1].function.column == 'b'
    assert args['whereclause'][1].value == '7'
    assert args['limit'] == d['limit']
    #assert args['order_by'] == d['order_by'] ##TODO
    assert args['plot'] == d['plot']
    assert args['modelids'] == d['modelids']
    assert args['summarize'] == d['summarize']
    assert args['newtablename'] == d['newtablename']

    methods, args, client_dict = parser.parse_single_statement(bql_statement.parseString('freq select a from t'))
    d = dict(tablename=tablename, plot=plot, modelids=None, summarize=False, hist=False, freq=True)
    assert method == 'select'
    assert args['tablename'] == d['tablename']
    assert args['plot'] == d['plot']
    assert args['modelids'] == d['modelids']
    assert args['summarize'] == d['summarize']
    assert args['freq'] == d['freq']
    assert args['hist'] == d['hist']

    methods, args, client_dict = parser.parse_single_statement(bql_statement.parseString('hist select a from t'))
    d = dict(tablename=tablename, plot=plot, modelids=None, summarize=False, hist=True, freq=False)
    assert method == 'select'
    assert args['tablename'] == d['tablename']
    assert args['plot'] == d['plot']
    assert args['modelids'] == d['modelids']
    assert args['freq'] == d['freq']
    assert args['hist'] == d['hist']

def test_infer(): ##TODO
    ##TODO test client_dict
    tablename = 't'
    newtablename = 'newtable'
    functions = function_in_query.parseString('*',parseAll=True)
    whereclause = None
    limit = float('inf')
    order_by = False
    plot = False

    method, args, client_dict = parser.parse_single_statement(bql_statement.parseString('infer * from t'))
    d = dict(tablename=tablename, functions=functions, whereclause=whereclause,
             limit=limit, order_by=order_by, plot=plot, modelids=None, summarize=False)
    assert method == 'infer'
    assert args['tablename'] == d['tablename']
    assert args['functions'][0].column_id == d['functions'][0].column_id
    assert args['whereclause'] == d['whereclause']
    assert args['limit'] == d['limit']
    assert args['order_by'] == d['order_by']
    assert args['plot'] == d['plot']
    assert args['modelids'] == d['modelids']
    assert args['summarize'] == d['summarize']

    method, args, client_dict = parser.parse_single_statement(bql_statement.parseString('summarize infer * from t'))

    d = dict(tablename=tablename, functions=functions, whereclause=whereclause,
             limit=limit, order_by=order_by, plot=plot, modelids=None, summarize=True)
    assert method == 'infer'
    assert args['tablename'] == d['tablename']
    assert args['functions'][0].column_id == d['functions'][0].column_id
    assert args['whereclause'] == d['whereclause']
    assert args['limit'] == d['limit']
    assert args['order_by'] == d['order_by']
    assert args['plot'] == d['plot']
    assert args['modelids'] == d['modelids']
    assert args['summarize'] == d['summarize']

    method, args, client_dict = parser.parse_single_statement(bql_statement.parseString('infer a, b, a_b from t'))
    d = dict(tablename=tablename, functions=None, whereclause=whereclause,
             limit=limit, order_by=order_by, plot=plot, modelids=None, summarize=False)
    assert method == 'infer'
    assert args['tablename'] == d['tablename']
    assert args['functions'][0].column_id == 'a'
    assert args['functions'][1].column_id == 'b'
    assert args['functions'][2].column_id == 'a_b'
    assert args['whereclause'] == d['whereclause']
    assert args['limit'] == d['limit']
    assert args['order_by'] == d['order_by']
    assert args['plot'] == d['plot']
    assert args['modelids'] == d['modelids']
    assert args['summarize'] == d['summarize']

    method, args, client_dict = parser.parse_single_statement(bql_statement.parseString('infer * from t where a=6 and b = 7'))
    d = dict(tablename=tablename, functions=functions, whereclause=whereclause,
             limit=limit, order_by=order_by, plot=plot, modelids=None, summarize=False)
    assert method == 'infer'
    assert args['tablename'] == d['tablename']
    assert args['functions'][0].column_id == '*'
    assert args['whereclause'][0].function.column == 'a'
    assert args['whereclause'][0].value == '6'
    assert args['whereclause'][1].function.column == 'b'
    assert args['whereclause'][1].value == '7'
    assert args['limit'] == d['limit']
    assert args['order_by'] == d['order_by']
    assert args['plot'] == d['plot']
    assert args['modelids'] == d['modelids']
    assert args['summarize'] == d['summarize']

    limit = 10
    method, args, client_dict = parser.parse_single_statement(bql_statement.parseString('infer * from t where a=6 and b = 7 limit 10'))
    d = dict(tablename=tablename, functions=functions, whereclause=whereclause,
             limit=limit, order_by=order_by, plot=plot, modelids=None, summarize=False)
    assert method == 'infer'
    assert args['tablename'] == d['tablename']
    assert args['functions'][0].column_id == d['functions'][0].column_id
    assert args['whereclause'][0].function.column == 'a'
    assert args['whereclause'][0].value == '6'
    assert args['whereclause'][1].function.column == 'b'
    assert args['whereclause'][1].value == '7'
    assert args['limit'] == d['limit']
    assert args['order_by'] == d['order_by']
    assert args['plot'] == d['plot']
    assert args['modelids'] == d['modelids']
    assert args['summarize'] == d['summarize']

    method, args, client_dict = parser.parse_single_statement(bql_statement.parseString('infer * from t where a=6 and b = 7 limit 10 into newtable'))
    d = dict(tablename=tablename, functions=functions, whereclause=whereclause,
             limit=limit, order_by=order_by, plot=plot, modelids=None, summarize=False, newtablename=newtablename)
    assert method == 'infer'
    assert args['tablename'] == d['tablename']
    assert args['functions'][0].column_id == d['functions'][0].column_id
    assert args['whereclause'][0].function.column == 'a'
    assert args['whereclause'][0].value == '6'
    assert args['whereclause'][1].function.column == 'b'
    assert args['whereclause'][1].value == '7'
    assert args['limit'] == d['limit']
    assert args['order_by'] == d['order_by']
    assert args['plot'] == d['plot']
    assert args['modelids'] == d['modelids']
    assert args['summarize'] == d['summarize']
    assert args['newtablename'] == d['newtablename']

    order_by = [('b', False)]
    method, args, client_dict = parser.parse_single_statement(bql_statement.parseString('infer * from t where a=6 and b = 7 order by b limit 10'))
    d = dict(tablename=tablename, functions=functions, whereclause=whereclause,
             limit=limit, order_by=order_by, plot=plot, modelids=None, summarize=False)
    assert method == 'infer'
    assert args['tablename'] == d['tablename']
    assert args['functions'][0].column_id == d['functions'][0].column_id
    assert args['whereclause'][0].function.column == 'a'
    assert args['whereclause'][0].value == '6'
    assert args['whereclause'][1].function.column == 'b'
    assert args['whereclause'][1].value == '7'
    assert args['limit'] == d['limit']
    #assert args['order_by'] == d['order_by'] ##TODO
    assert args['plot'] == d['plot']
    assert args['modelids'] == d['modelids']
    assert args['summarize'] == d['summarize']

def test_simulate(): ##TODO
    method, args, client_dict = parser.parse_single_statement(bql_statement.parseString('simulate * from t times 10'))
    assert method == 'simulate'
    assert args['tablename'] == 't'
    assert args['functions'][0].column_id == '*'
    assert args['summarize'] == False
    assert args['plot'] == False
    assert args['order_by'] == False
    assert args['modelids'] == None
    assert args['newtablename'] == None
    assert args['givens'] == None
    assert args['numpredictions'] == 10

    assert client_dict['pairwise'] == False
    assert client_dict['filename'] == None
    assert client_dict['scatter'] == False

    method, args, client_dict = parser.parse_single_statement(bql_statement.parseString('simulate * from t times 10 into newtable'))
    assert method == 'simulate'
    assert args['tablename'] == 't'
    assert args['functions'][0].column_id == '*'
    assert args['summarize'] == False
    assert args['plot'] == False
    assert args['order_by'] == False
    assert args['modelids'] == None
    assert args['newtablename'] == 'newtable'
    assert args['givens'] == None
    assert args['numpredictions'] == 10

    assert client_dict['pairwise'] == False
    assert client_dict['filename'] == None
    assert client_dict['scatter'] == False
    ##TODO more clauses

def test_estimate():
    method, args, client_dict = parser.parse_single_statement(bql_statement.parseString('estimate columns from t'))
    assert method == 'estimate_columns'
    assert args['tablename'] == 't'
    assert args['functions'][0] == 'column'
    assert args['whereclause'] == None
    assert args['limit'] == float('inf')
    assert args['order_by'] == False
    assert args['name'] == None
    assert args['modelids'] == None
    assert client_dict == None

def test_estimate_pairwise():
    method, args, client_dict = parser.parse_single_statement(bql_statement.parseString('estimate pairwise correlation from t'))
    assert method == 'estimate_pairwise'
    assert args['tablename'] == 't'
    assert args['function_name'] == 'correlation'
    assert args['column_list'] == None
    assert args['clusters_name'] == None
    assert args['threshold'] == None
    assert args['modelids'] == None
    assert client_dict['filename'] == None


def test_estimate_pairwise_row():
    method, args, client_dict = parser.parse_single_statement(bql_statement.parseString('estimate pairwise row similarity from t'))
    assert method == 'estimate_pairwise_row'
    assert args['tablename'] == 't'
    assert args['function'].function_id == 'similarity'
    assert args['row_list'] == None
    assert args['clusters_name'] == None
    assert args['threshold'] == None
    assert args['modelids'] == None
    assert client_dict['filename'] == None

def test_disallowed_queries():
    """
    All of these queries should pass the grammar and fail at parser.parse_query
    """
    strings = ["select * from test times 10",
               "select * from test save clusters with threshold .5 as test.csv",
               "select * from test given a=5",
               "select * from test with confidence .4",
               "select a conf .4 from test",
               "select a conf .4, b from test",
               "simulate a conf .4 from test times 10",
               "simulate a conf .4, b from test times 10",
               "infer * from test times 10",
               "infer typicality from test",
               "infer * from test with confidence 1.5",
               "simulate typicality from test",
               "infer * from test save clusters with threshold .5 as test.csv",
               "infer * from test given a=5",
               "simulate * from test where a < 4",
               "simulate * from test save clusters with threshold .5 as test.csv",
               "simulate * from test with confidence .4",
               "simulate * from test with 4 samples",
               "simulate * from test",
               "estimate columns from test with confidence .4",
               "estimate columns from test given a=4",
               "estimate columns from test times 10",
               "summarize estimate columns from test",
               "plot estimate columns from test",
               "estimate columns from test save clusters with threshold .5 as test.csv",
               "estimate pairwise correlation from test where a = b",
               "estimate pairwise correlation from test times 10",
               "estimate pairwise correlation from test given a = 5",
               "estimate pairwise correlation from test with confidence .2",
               "estimate pairwise row similarity from test where a = b",
               "estimate pairwise row similarity from test times 10",
               "estimate pairwise row similarity from test given a = 5",
               "estimate pairwise row similarity from test with confidence .2",
               "estimate pairwise row similarity from test where a = b"
               ]

    for query_string in strings:
        ast = bql_statement.parseString(query_string,parseAll=True)
        with pytest.raises(AssertionError):
            parser.parse_single_statement(ast)

def test_old_grammar_fails():
    """
    All of these queries are formerly valid but should not pass the current grammar.
    """

    strings = [
        'update schema for test set a = continuous',
        'update schema for test set a = multinomial',
        # Cyclic is invalid because it should be followed by (min, max) parameters.
        'update schema for test set a = cyclic'
    ]

    for query_string in strings:
        with pytest.raises(ParseException):
            ast = bql_statement.parseString(query_string, parseAll=True)

def test_label_and_metadata():
    # LABEL COLUMNS FOR <btable> SET <column1 = column-label-1> [, <column-name-2 = column-label-2>, ...]
    query_str1 = 'label columns for test set col_1 = label1, col_2 = "label 2"'
    ast = bql_statement.parseString(query_str1, parseAll=True)
    assert ast.label_clause[0][0] == 'col_1'
    assert ast.label_clause[0][1] == 'label1'
    assert ast.label_clause[1][0] == 'col_2'
    assert ast.label_clause[1][1] == 'label 2'



#
#   Copyright (c) 2010-2014, MIT Probabilistic Computing Project
#
#   Lead Developers: Jay Baxter and Dan Lovell
#   Authors: Jay Baxter, Dan Lovell, Baxter Eaves, Vikash Mansinghka
#   Research Leads: Vikash Mansinghka, Patrick Shafto
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from __future__ import print_function
#
#  Copyright (c) 2011 Edward Langley
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
#  are met:
#
#  Redistributions of source code must retain the above copyright notice,
#  this list of conditions and the following disclaimer.
#
#  Redistributions in binary form must reproduce the above copyright
#  notice, this list of conditions and the following disclaimer in the
#  documentation and/or other materials provided with the distribution.
#
#  Neither the name of the project's author nor the names of its
#  contributors may be used to endorse or promote products derived from
#  this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
#  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
#  HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
#  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
#  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
#  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
#  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#

from twisted.internet import ssl
import traceback

from twisted.internet import reactor
from twisted.web import server, iweb
from twisted.web.resource import EncodingResourceWrapper

from jsonrpc.server import ServerEvents, JSON_RPC

import bayesdb.engine as engine_module
engine_methods = engine_module.get_method_names()
from bayesdb.engine import Engine
engine = Engine()

class ExampleServer(ServerEvents):
      # inherited hooks
      def log(self, responses, txrequest, error):
            print(txrequest.code, end=' ')
            if isinstance(responses, list):
                  for response in responses:
                        msg = self._get_msg(response)
                        print(txrequest, msg)
            else:
                  msg = self._get_msg(responses)
                  print(txrequest, msg)
                        
      def findmethod(self, method, args=None, kwargs=None):
            if method in self.methods:
                  return getattr(engine, method)
            else:
                  return None
            
      # helper methods
      methods = set(engine_methods)
      def _get_msg(self, response):
            ret_str = 'No id response: %s' % str(response)
            if hasattr(response, 'id'):
                  ret_str = str(response.id)
                  if response.result:
                        ret_str += '; result: %s' % str(response.result)
                  else:
                        ret_str += '; error: %s' % str(response.error)
                        for at in dir(response):
                              if not at.startswith('__'):
                                    print(at + ": " + str(getattr(response, at)))
                        print("response:\n" + str(dir(response)))
            return ret_str
      
      
class CorsEncoderFactory(object):
      def encoderForRequest(self, request):
            request.setHeader("Access-Control-Allow-Origin", '*')
            request.setHeader("Access-Control-Allow-Methods", 'PUT, GET')
            return _CorsEncoder(request)

class _CorsEncoder(object):
      """
      @ivar _request: A reference to the originating request.
      
      @since: 12.3
      """
      
      def __init__(self, request):
            self._request = request
            
      def encode(self, data):
            return data
      
      def finish(self):
            return ""


root = JSON_RPC().customize(ExampleServer)
wrapped = EncodingResourceWrapper(root, [CorsEncoderFactory()])
site = server.Site(wrapped)

# 8008 is the port you want to run under. Choose something >1024
PORT = 8008
print('Listening on port %d...' % PORT)
reactor.listenTCP(PORT, site)
reactor.run()



#
#   Copyright (c) 2010-2014, MIT Probabilistic Computing Project
#
#   Lead Developers: Jay Baxter and Dan Lovell
#   Authors: Jay Baxter, Dan Lovell, Baxter Eaves, Vikash Mansinghka
#   Research Leads: Vikash Mansinghka, Patrick Shafto
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import inspect
import pickle
import gzip
import prettytable
import re
import os
import time
import ast
from textwrap import dedent

import utils
import data_utils
import plotting_utils
import api_utils
from parser import Parser
from engine import Engine


class Client(object):
    def __init__(self, crosscat_host=None, crosscat_port=8007,
                 crosscat_engine_type='multiprocessing', bayesdb_host=None, bayesdb_port=8008,
                 seed=None, testing=False):
        """
        Create a client object. The client creates a parser, that is uses to parse all commands,
        and an engine, which is uses to execute all commands. The engine can be remote or local.
        If local, the engine will be created.
        """
        self.parser = Parser()
        if bayesdb_host is None or bayesdb_host == 'localhost':
            self.online = False
            self.engine = Engine(crosscat_host, crosscat_port, crosscat_engine_type, seed,
                                 testing=testing)
        else:
            self.online = True
            self.hostname = bayesdb_host
            self.port = bayesdb_port
            self.URI = 'http://' + self.hostname + ':%d' % self.port

    def call_bayesdb_engine(self, method_name, args_dict, debug=False):
        """
        Helper function used to call the BayesDB engine, whether it is remote or local.
        Accepts method name and arguments for that method as input.
        If this is the first time a query has been executed on the given tablename
        during this session, first check that the table doesn't need any upgrades to
        work with the current version of BayesDB. (After checking/upgrading, the table
        will be considered upgraded for the duration of the session.)
        """
        table_query = 'tablename' in args_dict
        if table_query:
            tablename = args_dict['tablename']
            if self.online:
                out, id = api_utils.call('check_btable_created_and_checked',
                                         {'tablename': tablename, 'client_online': True}, self.URI)
                table_created = out['result']['table_created']
                table_checked = out['result']['table_checked']
            else:
                table_created, table_checked = self.engine.check_btable_created_and_checked(tablename)
        if self.online:
            if table_query:
                if table_created and not table_checked and method_name is not 'drop_btable':
                    out, id = aqupi_utils.call('upgrade_btable', {'tablename': tablename}, self.URI)
            out, id = aqupi_utils.call(method_name, args_dict, self.URI)
        else:
            if table_query:
                if table_created and not table_checked and method_name is not 'drop_btable':
                    self.engine.upgrade_btable(tablename)
            method = getattr(self.engine, method_name)
            if debug:
                out = method(**args_dict)
            else:
                # when not in debug mode, catch all BayesDBErrors
                try:
                    out = method(**args_dict)
                except utils.BayesDBError as e:
                    out = dict(message=str(e), error=True)
        return out

    def __call__(self, call_input, pretty=True, timing=False, wait=False, plots=None, yes=False,
                 debug=False, pandas_df=None, pandas_output=True, key_column=None,
                 return_raw_result=False, force_output=False):
        """Wrapper around execute."""
        return self.execute(call_input, pretty, timing, wait, plots, yes, debug, pandas_df,
                            pandas_output, key_column, return_raw_result, force_output)

    def execute(self, call_input, pretty=True, timing=False, wait=False, plots=None, yes=False,
                debug=False, pandas_df=None, pandas_output=True, key_column=None,
                return_raw_result=False, force_output=False):
        """
        Execute a chunk of BQL. This method breaks a large chunk of BQL (like a file)
        consisting of possibly many BQL statements, breaks them up into individual statements,
        then passes each individual line to self.execute_statement() as a string.

        param call_input: may be either a file object, or a string.
        If the input is a file, then we load the inputs of the file, and use those as a string.

        See self.execute_statement() for an explanation of arguments.
        """
        if type(call_input) == file:
            bql_string = call_input.read()
            path = os.path.abspath(call_input.name)
            self.parser.set_root_dir(os.path.dirname(path))
        elif type(call_input) == str:
            bql_string = call_input
        else:
            try:
                call_input.encode('ascii', 'ignore')
                bql_string = call_input
            except:
                raise ValueError("Invalid input type: expected file or string. Got: %s of type %s."
                                 % (call_input, type(call_input)))

        return_list = []

        # Parse input, but catch parsing errors and abort
        try:
            lines = [bql_statement_ast for bql_statement_ast in self.parser.pyparse_input(bql_string)]
        except utils.BayesDBError as e:
            if debug:
                raise e
            else:
                print(str(e))
                return

        # Iterate through lines with while loop so we can append within loop.
        while len(lines) > 0:
            line = lines.pop(0)
            if type(call_input) == file:
                print('> %s' % line)
            if wait:
                user_input = raw_input()
                if len(user_input) > 0 and (user_input[0] == 'q' or user_input[0] == 's'):
                    continue
            result = self.execute_statement(line, pretty=pretty, timing=timing, plots=plots,
                                            yes=yes, debug=debug, pandas_df=pandas_df,
                                            pandas_output=pandas_output, key_column=key_column,
                                            return_raw_result=return_raw_result,
                                            force_output=force_output)

            if type(result) == dict and 'message' in result and result['message'] == 'execute_file':
                # special case for one command: execute_file
                new_lines = self.parser.split_lines(result['bql_string'])
                lines += new_lines
            if type(call_input) == file:
                print

            return_list.append(result)

        self.parser.reset_root_dir()

        if not pretty or return_raw_result or force_output:
            return return_list

    def execute_statement(self, bql_statement_ast, pretty=True, timing=False, plots=None, yes=False,
                          debug=False, pandas_df=None, pandas_output=True, key_column=None,
                          return_raw_result=False, force_output=False):
        """
        Accepts a SINGLE BQL STATEMENT as input, parses it, and executes it if it was parsed
        successfully.

        If pretty=True, then the command output will be pretty-printed as a string.
        If pretty=False, then the command output will be returned as a python object.
        If force_output=True, then results will be returned regardless of pretty

        timing=True prints out how long the command took to execute.

        For commands that have visual results, plots=True will cause those to be displayed
        by matplotlib as graphics rather than being pretty-printed as text.
        (Note that the graphics will also be saved if the user added SAVE TO <filename> to the BQL.)
        """
        if timing:
            start_time = time.time()

        parser_out = None
        # TODO move pyparsing objects out of client into parser
        if debug:
            parser_out = self.parser.parse_single_statement(bql_statement_ast)
        else:
            try:
                parser_out = self.parser.parse_single_statement(bql_statement_ast)
            except Exception as e:
                raise utils.BayesDBParseError(str(e))
        if parser_out is None:
            print("Could not parse command. Try typing 'help' for a list of all commands.")
            return
        elif not parser_out:
            return

        method_name, args_dict, client_dict = parser_out
        if client_dict is None:
            client_dict = {}

        # Do stuff now that you know the user's command, but before passing it to engine.
        if method_name == 'execute_file':
            return dict(message='execute_file', bql_string=open(args_dict['filename'], 'r').read())
        elif method_name == 'update_codebook':
            _, codebook_rows = data_utils.read_csv(client_dict['codebook_path'], has_header=True)
            # TODO: require specific codebook_header values? Or don't require a header,
            # and if the first value in the header is actually a data column name, assume
            # the first row is codebook data, not a header.

            # Create a dict indexed by column name
            codebook = dict()
            for codebook_row in codebook_rows:
                codebook[codebook_row[0]] = dict(zip(['short_name', 'description', 'value_map'],
                                                 codebook_row[1:]))

            args_dict['codebook'] = codebook
        elif (method_name == 'drop_btable') and (not yes):
            # If dropping something, ask for confirmation.
            print("Are you sure you want to permanently delete this btable, and all associated "
                  "models, without any way to get them back? Enter 'y' if yes.")
            user_confirmation = raw_input()
            if 'y' != user_confirmation.strip():
                return dict(message="Operation canceled by user.")
        elif (method_name == 'drop_models') and (not yes):
            # If dropping something, ask for confirmation.
            print("Are you sure you want to permanently delete model(s), without any way to get "
                  "them back? Enter 'y' if yes.")
            user_confirmation = raw_input()
            if 'y' != user_confirmation.strip():
                return dict(message="Operation canceled by user.")
        elif method_name == 'load_models':
            pklpath = client_dict['pkl_path']
            try:
                model_data = pickle.load(gzip.open(self.parser.get_absolute_path(pklpath), 'rb'))
            except IOError as e:
                if pklpath[-7:] != '.pkl.gz':
                    if pklpath[-4:] == '.pkl':
                        model_data = pickle.load(open(self.parser.get_absolute_path(pklpath), 'rb'))
                    else:
                        pklpath = pklpath + ".pkl.gz"
                        model_data = pickle.load(gzip.open(self.parser.get_absolute_path(pklpath),
                                                 'rb'))
                else:
                    raise utils.BayesDBError('Models file %s could not be found.' % pklpath)
            # This is the more recent version, where schema is stored with models.
            if 'schema' in model_data.keys():
                args_dict['models'] = model_data['models']
                args_dict['model_schema'] = model_data['schema']
            # This support older saved models, where only the model info was stored.
            else:
                args_dict['models'] = model_data
                args_dict['model_schema'] = None

            # Older versions of model_schema just had a str cctype as the dict items.
            # Newest version has a dict of cctype and parameters. Use this values to
            # test the recency of the models.
            model_schema = args_dict['model_schema']
            if model_schema:
                model_schema_itemtype = type(model_schema[model_schema.keys()[0]])
            else:
                model_schema_itemtype = None

            if model_schema is None or model_schema_itemtype != dict:
                args_dict['model_schema'] = None
                if not yes:
                    print """WARNING! The models you are currently importing were saved without a schema
                        or without detailed column parameters (probably from a previous version).

                        If you are loading models into the same table from which you created them, problems
                        are unlikely, unless you have dropped models and then updated the schema.

                        If you are loading models into a different table from which you created them, you
                        should verify that the table schemas are the same.

                        Please use "SAVE MODELS FROM <btable> TO <filename.pkl.gz>" to create an updated copy of your models.

                        Are you sure you want to load these model(s)?
                        """
                    user_confirmation = raw_input()
                    if 'y' != user_confirmation.strip():
                        return dict(message="Operation canceled by user.")
        elif method_name == 'create_btable':
            if pandas_df is None:
                header, rows = data_utils.read_csv(client_dict['csv_path'])
            else:
                header, rows = data_utils.read_pandas_df(pandas_df)
            args_dict['header'] = header
            args_dict['raw_T_full'] = rows
            args_dict['key_column'] = key_column
            args_dict['subsample'] = False

            if 'codebook_path' in client_dict:
                _, codebook_rows = data_utils.read_csv(client_dict['codebook_path'],
                                                       has_header=True)
                # TODO: require specific codebook_header values? Or don't require a header,
                # and if the first value in the header is actually a data column name, assume
                # the first row is codebook data, not a header.

                # Create a dict indexed by column name
                codebook = dict()
                for codebook_row in codebook_rows:
                    codebook[codebook_row[0]] = dict(zip(['short_name', 'description', 'value_map'],
                                                     codebook_row[1:]))
                args_dict['codebook'] = codebook
            else:
                warning = dedent("""
                WARNING!

                You are creating a btable without a codebook, which will make interpretation
                of results more difficult. Codebooks should be in CSV format with each row
                corresponding to one column of the original data. The codebook should have four
                columns:

                1. actual column name
                2. short column description
                3. long column description
                4. value map (optional, only used for categorical columns - should be in JSON
                   format)
                """)
                print(warning)

            # Display warning messages and get confirmation if btable is too large.
            # Ask user if they want to turn on subsampling.
            max_columns = 200
            max_rows = 1000
            max_cells = 100000
            message = None
            if not yes:
                if len(rows[0]) > max_columns:
                    message = "The btable you are uploading has %d columns, but BayesDB is " \
                              "currently designed to support only %d columns. If you proceed, " \
                              "performance may suffer unless you set many columns' datatypes to " \
                              "'ignore'. Would you like to continue? Enter 'y' if yes." \
                              % (len(rows[0]), max_columns)
                if len(rows) > max_rows:
                    message = "The btable you are uploading has %d rows, but BayesDB is currently "\
                              "designed to support only %d rows. If you proceed, performance may "\
                              "suffer. Would you like to continue? Enter 'y' to continue without "\
                              "subsampling, 'n' to abort, 's' to continue by subsampling %d rows, "\
                              "or a positive integer to specify the number of rows to be "\
                              "subsampled." % (len(rows), max_rows, max_rows)
                if len(rows[0])*len(rows) > max_cells:
                    message = "The btable you are uploading has %d cells, but BayesDB is currently"\
                              " designed to support only %d cells. If you proceed, performance may"\
                              " suffer unless you enable subsampling. Enter 'y' to continue "\
                              " without subsampling, 'n' to abort, 's' to continue by subsampling "\
                              "%d rows, or a positive integer to specify the number of rows to be "\
                              "subsampled." % (len(rows)*len(rows[0]), max_cells, max_rows)
                if message is not None:
                    print(message)
                    user_confirmation = raw_input()
                    if 'y' == user_confirmation.strip():
                        pass
                    elif 'n' == user_confirmation.strip():
                        return dict(message="Operation canceled by user.")
                    elif 's' == user_confirmation.strip():
                        args_dict['subsample'] = min(max_rows, len(rows))
                    elif utils.is_int(user_confirmation.strip()):
                        args_dict['subsample'] = int(user_confirmation.strip())
                    else:
                        return dict(message="Operation canceled by user.")
        elif method_name in ['label_columns', 'update_metadata']:
            if client_dict['source'] == 'file':
                header, rows = data_utils.read_csv(client_dict['csv_path'])
                args_dict['mappings'] = {key: value for key, value in rows}

        # Call engine.
        result = self.call_bayesdb_engine(method_name, args_dict, debug)

        # If error occurred, exit now.
        if 'error' in result and result['error']:
            if pretty:
                print(result['message'])
                if force_output:
                    return result
                else:
                    return result['message']
            else:
                return result

        # Do stuff now that engine has given you output, but before printing the result.
        result = self.callback(method_name, args_dict, client_dict, result)

        if return_raw_result:
            raw_result = {
                'result': result,
                'method_name': method_name,
                'client_dict': client_dict}
            print("returning raw result for %s" % (method_name))
            return raw_result

        assert type(result) != int

        if timing:
            end_time = time.time()
            print('Elapsed time: %.2f seconds.' % (end_time - start_time))

        if plots is None:
            plots = 'DISPLAY' in os.environ.keys()

        if 'matrix' in result and (plots or client_dict['filename']):
            # Plot matrices
            plotting_utils.plot_matrix(result['matrix'], result['column_names'], result['title'],
                                       client_dict['filename'])
            if pretty:
                if 'column_lists' in result:
                    print(self.pretty_print(dict(column_lists=result['column_lists'])))

                if force_output:
                    return result
                else:
                    return self.pretty_print(result)
            else:
                return result
        if ('plot' in client_dict and client_dict['plot']):
            if (plots or client_dict['filename']):
                # Plot generalized histograms or scatterplots

                try:
                    plotting_M_c = result['metadata_full']['M_c_full']
                except KeyError:
                    plotting_M_c = result['M_c']

                plot_remove_key = method_name in ['select', 'infer']
                plotting_utils.plot_general_histogram(result['column_names'], result['data'],
                                                      plotting_M_c, result['schema_full'],
                                                      client_dict['filename'],
                                                      client_dict['scatter'],
                                                      remove_key=plot_remove_key)
                return self.pretty_print(result)
            else:
                if 'message' not in result:
                    result['message'] = ""
                result['message'] = "Your query indicates that you would like to make a plot, but "\
                                    "in order to do so, you must either enable plotting in a "\
                                    "window or specify a filename to save to by appending 'SAVE "\
                                    "TO <filename>' to this command.\n" + result['message']

        if pretty:
            pp = self.pretty_print(result)
            print(pp)

        # Print warnings last so they're readable without scrolling backwards.
        if 'warnings' in result:
            """ Pretty-print warnings. """
            for warning in result['warnings']:
                print('WARNING: %s' % warning)

        if pandas_output and 'data' in result and 'column_labels' in result:
            result_pandas_df = data_utils.construct_pandas_df(result)
            return result_pandas_df
        else:
            return result

    def callback(self, method_name, args_dict, client_dict, result):
        """
        This method is meant to be called after receiving the result of a
        call to the BayesDB engine, and modifies the output before it is displayed
        to the user.
        """
        if method_name == 'save_models':
            samples_dict = result
            # Here is where the models get saved.
            pkl_path = client_dict['pkl_path']
            if pkl_path[-7:] != '.pkl.gz':
                if pkl_path[-4:] == '.pkl':
                    pkl_path = pkl_path + ".gz"
                else:
                    pkl_path = pkl_path + ".pkl.gz"
            samples_file = gzip.GzipFile(pkl_path, 'wb')
            pickle.dump(samples_dict, samples_file)
            return dict(message="Successfully saved the samples to %s" % client_dict['pkl_path'])

        else:
            return result

    def pretty_print(self, query_obj):
        """
        Return a pretty string representing the output object.
        """
        assert type(query_obj) == dict
        result = ""
        if type(query_obj) == dict and 'message' in query_obj:
            result += query_obj["message"] + "\n"
        if 'data' in query_obj and 'column_labels' in query_obj:
            """ Pretty-print data table """
            pt = prettytable.PrettyTable()
            columns = query_obj['column_labels']
            pt.field_names = columns

            # Adjust value width - for now preserve 2 decimal places.
            for row_idx, row_values in enumerate(query_obj['data']):
                if type(row_values) == tuple:
                    row_values = list(row_values)
                for col_idx, col_value in enumerate(row_values):
                    if type(col_value) == float:
                        # Right-align numeric columns.
                        if row_idx == 0:
                            pt.align[columns[col_idx]] = 'r'
                        row_values[col_idx] = "% .2f" % col_value
                pt.add_row(row_values)
            result += str(pt)

        elif 'list' in query_obj:
            """ Pretty-print lists """
            result += str(query_obj['list'])
        elif 'column_names' in query_obj:
            """ Pretty-print cctypes """
            colnames = query_obj['column_names']
            zmatrix = query_obj['matrix']
            pt = prettytable.PrettyTable(hrules=prettytable.ALL, vrules=prettytable.ALL,
                                         header=False)
            pt.add_row([''] + list(colnames))
            for row, colname in zip(zmatrix, list(colnames)):
                pt.add_row([colname] + list(row))
            result += str(pt)
        elif 'column_labels' in query_obj:
            """ Pretty-print column list."""
            pt = prettytable.PrettyTable()
            pt.field_names = ['column']
            for column in query_obj['column_labels']:
                pt.add_row([column])
            result += str(pt)
        elif 'row_lists' in query_obj:
            """ Pretty-print multiple row lists, which are just names and row sizes. """
            pt = prettytable.PrettyTable()
            pt.field_names = ('Row List Name', 'Row Count')

            def get_row_list_sorting_key(x):
                """ To be used as the key function in a sort. Puts cc_2 ahead of cc_10, e.g. """
                name, count = x
                if '_' not in name:
                    return name
                s = name.split('_')
                end = s[-1]
                start = '_'.join(s[:-1])
                if utils.is_int(end):
                    return (start, int(end))
                return name

            for name, count in sorted(query_obj['row_lists'], key=get_row_list_sorting_key):
                pt.add_row((name, count))
            result += str(pt)
        elif 'column_lists' in query_obj:
            """ Pretty-print multiple column lists. """
            print
            clists = query_obj['column_lists']
            for name, clist in clists:
                print("%s:" % name)
                pt = prettytable.PrettyTable()
                pt.field_names = clist
                print(pt)
        elif 'models' in query_obj:
            """ Pretty-print model info. """
            pt = prettytable.PrettyTable()
            pt.field_names = ('model_id', 'iterations')
            for (id, iterations) in query_obj['models']:
                pt.add_row((id, iterations))
            result += str(pt)

        if len(result) >= 1 and result[-1] == '\n':
            result = result[:-1]
        return result



#
#   Copyright (c) 2010-2014, MIT Probabilistic Computing Project
#
#   Lead Developers: Jay Baxter and Dan Lovell
#   Authors: Jay Baxter, Dan Lovell, Baxter Eaves, Vikash Mansinghka
#   Research Leads: Vikash Mansinghka, Patrick Shafto
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import inspect
import numpy
import os
import re
import inspect
import ast
import pylab
import matplotlib.cm
import time
import pandas
import math
from collections import Counter

import data_utils as du
import select_utils
import functions

class BayesDBError(Exception):
    """ Base class for all other exceptions in this module. """
    pass

class BayesDBParseError(BayesDBError):
    def __init__(self, msg=None):
        if msg:
            self.msg = msg
        else:
            self.msg = "BayesDB parsing error. Try using 'help' to see the help menu for BQL syntax."

    def __str__(self):
        return self.msg

class BayesDBUniqueValueError(BayesDBError):
    def __init__(self, msg=None):
        if msg:
            self.msg = msg
        else:
            self.msg = "BayesDB unique value error. More than one row has this value."

    def __str__(self):
        return self.msg


class BayesDBNoModelsError(BayesDBError):
    def __init__(self, tablename):
        self.tablename = tablename

    def __str__(self):
        return "Btable %s has no models, but this command requires models. Please create models first with INITIALIZE MODELS, and then ANALYZE." % self.tablename

class BayesDBInvalidBtableError(BayesDBError):
    def __init__(self, tablename):
        self.tablename = tablename

    def __str__(self):
        return "Btable %s does not exist. Please create it first with CREATE BTABLE, or view existing btables with LIST BTABLES." % self.tablename

class BayesDBColumnDoesNotExistError(BayesDBError):
    def __init__(self, column, tablename):
        self.column = column
        self.tablename = tablename

    def __str__(self):
        return "Column %s does not exist in btable %s." % (self.column, self.tablename)

class BayesDBColumnListDoesNotExistError(BayesDBError):
    def __init__(self, column_list, tablename):
        self.column_list = column_list
        self.tablename = tablename

    def __str__(self):
        return "Column list %s does not exist in btable %s." % (self.column_list, self.tablename)

class BayesDBRowListDoesNotExistError(BayesDBError):
    def __init__(self, row_list, tablename):
        self.row_list = row_list
        self.tablename = tablename

    def __str__(self):
        return "Row list %s does not exist in btable %s." % (self.row_list, self.tablename)

def is_int(s):
    try:
        int(s)
        return True
    except ValueError:
        return False

def is_float(s):
    try:
        float(s)
        return True
    except ValueError:
        return False

def value_string_to_num(value_string):
    if is_int(value_string) == True:
        value = int(value_string)
    elif is_float(value_string) == True:
        value = float(value_string)
    else:
        raise BayesDBParseError("Number expected for value: %s" % value_string)
    return value

def string_to_column_type(value_string, column, M_c):
    """
    column is the string of the column name
    Checks the type of the column in question based on M_c
    If numerical, converts the value from string to int or float
    """
    value = value_string
    if get_cctype_from_M_c(M_c, column) in ['numerical', 'cyclic']:
        if is_int(value_string) == True:
            value = int(value)
        elif is_float(value_string) == True:
            value = float(value)
    return value

def row_id_from_col_value(value, column, M_c, T):
    """
    Returns the row_id of a column where column == value
    If duplicate rows are found, raises exception
    If no rows are found, returns None
    """
    target_row_id = None
    col_idx = M_c['name_to_idx'][column]
    if type(value) == str:
        value = string_to_column_type(value, column, M_c)
    for row_id, T_row in enumerate(T):
        row_values = select_utils.convert_row_from_codes_to_values(T_row, M_c)
        if row_values[col_idx] == value:
            if target_row_id == None:
                target_row_id = row_id
            else:
                raise BayesDBUniqueValueError("Invalid Query: column '%s' has more than one row with value '%s'." %(column, str(value)))
    return target_row_id

##TODO move to engine
def infer(M_c, X_L_list, X_D_list, Y, row_id, col_id, numsamples, confidence, engine):
    q = [row_id, col_id]
    out = engine.call_backend('impute_and_confidence', dict(M_c=M_c, X_L=X_L_list, X_D=X_D_list, Y=Y, Q=[q], n=numsamples))
    code, conf = out
    if conf >= confidence:
      return code
    else:
      return None

def get_imputation_and_confidence_from_samples(M_c, X_L, col_idx, samples):
    samples = numpy.array(samples).T[0]
    modeltype = M_c['column_metadata'][col_idx]['modeltype']
    imputation_function = modeltype_to_imputation_function[modeltype]
    e = imputation_function(samples)
    imputation_confidence_function = modeltype_to_imputation_confidence_function[modeltype]
    column_component_suffstats_i = \
        get_column_component_suffstats_i(M_c, X_L, col_idx)
    imputation_confidence = \
        imputation_confidence_function(samples, e,
                                       column_component_suffstats_i)
    return e, imputation_confidence

    return lambda in_dict: in_dict[name]

def get_column_std(column_component_suffstats_i):
    N = sum(map(lambda in_dict: in_dict['N'], column_component_suffstats_i))
    sum_x = sum(map(lambda in_dict: in_dict['sum_x'], column_component_suffstats_i))
    sum_x_squared = sum(map(lambda in_dict: in_dict['sum_x_squared'], column_component_suffstats_i))
    #
    exp_x = sum_x / float(N)
    exp_x_squared = sum_x_squared / float(N)
    col_var = exp_x_squared - (exp_x ** 2)
    col_std = col_var ** .5
    return col_std


def get_column_component_suffstats_i(M_c, X_L, col_idx):
    column_name = M_c['idx_to_name'][str(col_idx)]
    view_idx = X_L['column_partition']['assignments'][col_idx]
    view_state_i = X_L['view_state'][view_idx]
    local_col_idx = view_state_i['column_names'].index(column_name)
    column_component_suffstats_i = \
        view_state_i['column_component_suffstats'][local_col_idx]
    return column_component_suffstats_i

def numerical_imputation(samples):
    imputed = numpy.median(samples)
    return imputed

def categorical_imputation(samples):
    counter = Counter(samples)
    max_tuple = counter.most_common(1)[0]
    max_count = max_tuple[1]
    counter_counter = Counter(counter.values())
    num_max_count = counter_counter[max_count]
    imputed = max_tuple[0]
    if num_max_count >= 1:
        # if there is a tie, draw randomly
        max_tuples = counter.most_common(num_max_count)
        values = [max_tuple[0] for max_tuple in max_tuples]
        random_state = numpy.random.RandomState()
        draw = random_state.randint(len(values))
        imputed = values[draw]
    return imputed

def categorical_imputation_confidence(samples, imputed, column_hypers_i):
    max_count = sum(numpy.array(samples) == imputed)
    confidence = float(max_count) / len(samples)
    return confidence

def get_numerical_mass_within_delta(samples, center, delta):
    num_samples = len(samples)
    num_within_delta = sum(numpy.abs(samples - center) < delta)
    mass_fraction = float(num_within_delta) / num_samples
    return mass_fraction

def numerical_imputation_confidence(samples, imputed,
                                     column_component_suffstats_i):
    col_std = get_column_std(column_component_suffstats_i)
    delta = .1 * col_std
    confidence = get_numerical_mass_within_delta(samples, imputed, delta)
    return confidence


modeltype_to_imputation_function = {
    'normal_inverse_gamma': numerical_imputation,
    'symmetric_dirichlet_discrete': categorical_imputation,
    }

modeltype_to_imputation_confidence_function = {
    'normal_inverse_gamma': numerical_imputation_confidence,
    'symmetric_dirichlet_discrete': categorical_imputation_confidence,
    }


def check_for_duplicate_columns(column_names):
    column_names_set = set()
    for name in column_names:
        if name in column_names_set:
            raise BayesDBError("Error: Column list has duplicate entries of column: %s" % name)
        column_names_set.add(name)


def get_all_column_names_in_original_order(M_c):
    colname_to_idx_dict = M_c['name_to_idx']
    colnames = map(lambda tup: tup[0], sorted(colname_to_idx_dict.items(), key=lambda tup: tup[1]))
    return colnames

def get_cctype_from_M_c(M_c, column):
    if column in M_c['name_to_idx'].keys():
        column_index = M_c['name_to_idx'][column]
        modeltype = M_c['column_metadata'][column_index]['modeltype']
        if modeltype == 'normal_inverse_gamma':
            cctype = 'numerical'
        elif modeltype == 'vonmises':
            cctype = 'cyclic'
        else:
            cctype = 'categorical'
    else:
        # If the column name wasn't found in metadata, it's a function, so the output will be numerical
        cctype = 'numerical'
    return cctype

# The 'inplace' argument to df.drop() was added to pandas in a version (which one??) that many people may
# not have. So, check to see if 'inplace' exists, otherwise don't pass it -- this just copies the dataframe.
def df_drop(df, column_list, **kwargs):
    if 'inplace' in inspect.getargspec(df.drop).args:
        df.drop(column_list, inplace=True, **kwargs)
    else:
        df = df.drop(column_list, **kwargs)


def summarize_freqs(x, n=5):
    """
    Function to return most frequent n values of each column of the DataFrame being summarized.
    Input: a DataFrame column, by default as Series type
    Input: n, the number of most common values to return (n=5 would be top 5 common values)

    Return: most frequent n values in x.
    """
    x_values, x_freqs, x_probs = get_column_freqs(x)

    if n is not None and len(x_values) > n:
        x_probs = x_probs[:n]
        x_values = x_values[:n]

    # Create index labels ('mode1/2/3/... and prob_mode1/2/3...')
    x_range = range(1, len(x_values) + 1)
    x_index = ['mode' + str(i) if i > 1 else 'mode' for i in x_range]
    x_index += ['prob_mode' + str(i) if i > 1 else 'prob_mode' for i in x_range]

    # Combine values and probabilities into a single list
    x_values.extend(x_probs)

    return pandas.Series(data = x_values, index = x_index)

def freqs(x):
    x_values, x_freqs, x_probs = get_column_freqs(x)
    x_values = pandas.Series(x_values)
    x_probs = pandas.Series(x_probs)
    x_freqs = pandas.Series(x_freqs)

    x_stats = pandas.concat([x_values, x_freqs, x_probs], axis=1)
    return x_stats

def get_column_freqs(x):
    """
    Takes a pandas.Series of data and returns
    1. the unique values
    2. the empirical probabilities of those values
    3. the empirical frequencies of those values
    """
    x_freqs  = x.value_counts()
    x_probs  = list(x_freqs / len(x))
    x_values = list(x_freqs.index)
    x_freqs  = list(x_freqs)

    return x_values, x_freqs, x_probs

def freq_table(data, columns, M_c, remove_key=True):
    """
    Returns a frequency table
    """
    if len(data) > 0:
        # Construct a pandas.DataFrame out of data and columns
        df = pandas.DataFrame(data=data, columns=columns)

        # If the first column is the key (select/infer), summarizing it is meaningless
        if remove_key:
            df_drop(df, [columns[0]], axis=1)

        column = df.columns[0]
        summary_data = freqs(df[column])
        columns = [column, 'frequency', 'probability']
        data = summary_data.to_records(index=False)

    return data, columns

def histogram_table(data, columns, M_c, remove_key=True):
    """
    Returns a frequency table
    """
    if len(data) > 0:
        # Construct a pandas.DataFrame out of data and columns
        df = pandas.DataFrame(data=data, columns=columns)

        # If the first column is the key (select/infer), summarizing it is meaningless
        if remove_key:
            df_drop(df, [columns[0]], axis=1)

        column = df.columns[0]
        
        # numpy.histogram can't deal with NaNs, so block those out
        col_data_no_nans = df[column].dropna()

        if get_cctype_from_M_c(M_c, column) in ['numerical', 'cyclic']:
            # Use Sturges formula to calculate the number of bins to use.
            n_bins = math.ceil(math.log(len(col_data_no_nans), 2) + 1)

            hist_data = numpy.histogram(col_data_no_nans, bins = n_bins)
            bin_mins = hist_data[1][:-1]
            bin_maxs = hist_data[1][1:]
            bin_freqs = hist_data[0]
            bin_probs = bin_freqs / float(sum(bin_freqs))
        else:
            hist_data = col_data_no_nans.value_counts().sort_index()
            bin_mins = list(hist_data.index.values)
            bin_maxs = list(hist_data.index.values)
            bin_freqs = list(hist_data)
            bin_probs = [ float(i)/sum(bin_freqs) for i in bin_freqs ]
            

        summary_data = pandas.DataFrame({
            'bin_minimum': bin_mins,
            'bin_maximum': bin_maxs,
            'frequency'  : bin_freqs,
            'probability': bin_probs
        })

        # Have to reorder columns, otherwise pandas defaults to alphabetical order
        columns = ['bin_minimum', 'bin_maximum', 'frequency', 'probability']
        summary_data = summary_data[columns]
        data = summary_data.to_records(index=False)

    return data, columns

def get_index_from_colname(M_c, column):
    if column in M_c['name_to_idx'].keys():
        return M_c['name_to_idx'][column]

def generate_pairwise_matrix(col_function_name, X_L_list, X_D_list, M_c, T, tablename='', confidence=None, limit=None, engine=None, column_names=None, component_threshold=None):
    """ Compute a matrix. If using a function that requires engine (currently only
    mutual information), engine must not be None. """

    # Get appropriate function
    assert len(X_L_list) == len(X_D_list)
    if col_function_name == 'mutual information':
      if len(X_L_list) == 0:
        return {'message': 'You must initialize models before computing mutual information.'}
      col_function = functions._mutual_information
    elif col_function_name == 'dependence probability':
      if len(X_L_list) == 0:
        return {'message': 'You must initialize models before computing dependence probability.'}
      col_function = functions._dependence_probability
    elif col_function_name == 'correlation':
      col_function = functions._correlation
    else:
      raise BayesDBParseError('Invalid column function: %s' % col_function_name)

    # If using a subset of the columns, get the appropriate names, and figure out their indices.
    if column_names is not None:
        num_cols = len(column_names)
        column_indices = [M_c['name_to_idx'][name] for name in column_names]
    else:
        BayesDBParseError("Invalid query: column '%s' not found" % column)

def summarize_table(data, columns, M_c, remove_key=True):
    """
    Returns a summary of the data.
    Input: data is a list of lists, of raw data values about to be shown to the user.
    Input: columns is a list of column names, as they will be displayed to the user. Note
    that some column names may be things like "row_id" or predictive functions, not actually
    columns.

    Return: columns should be the same, except with another column prepended called like "summaries" or something.
    Return: data should be summaries now.
    """
    if len(data) > 0:
        # Construct a pandas.DataFrame out of data and columns
        df = pandas.DataFrame(data=data, columns=columns)

        # If the first column is the key (select/infer), summarizing it is meaningless
        if remove_key:
            df_drop(df, [columns[0]], axis=1)

        # Get column types as one-row DataFrame
        cctypes = pandas.DataFrame([[get_cctype_from_M_c(M_c, col) for col in df.columns]], columns=df.columns, index=['type'])

        # Run pandas.DataFrame.describe() on each column - it'll compute every stat that it can for each column,
        # depending on its type (assume it's not a problem to overcompute here - for example, computing a mean on a
        # discrete variable with numeric values might not have meaning, but it's easier just to do it and
        # leave interpretation to the user, rather than try to figure out what's meaningful, especially with
        # columns that are the result of predictive functions.
        summary_describe = df.apply(pandas.Series.describe)

        # If there were discrete columns, remove 'top' and 'freq' rows, because we'll replace those
        # with the mode and empirical probabilities
        if 'top' in summary_describe.index and 'freq' in summary_describe.index:
            summary_describe = summary_describe.drop(['top', 'freq'])

        # Past versions used n=5 for 5 most frequent values, but now we have FREQ SELECT for freq tables.
        summary_freqs = df.apply(summarize_freqs, n=1)

        # Attach numerical and discrete summaries along row axis (unaligned values will be assigned NaN)
        summary_data = pandas.concat([cctypes, summary_describe, summary_freqs], axis=0)

        # Reorder rows: count, unique, mean, std, min, 25%, 50%, 75%, max, modes, prob_modes
        if hasattr(summary_data, 'loc'):
            potential_index = pandas.Index(['type', 'count', 'unique', 'mean', 'std', 'min', '25%', '50%', '75%', 'max', \
                'mode', 'mode2', 'mode3', 'mode4', 'mode5', \
                'prob_mode', 'prob_mode2', 'prob_mode3', 'prob_mode4', 'prob_mode5'])

            reorder_index = potential_index[potential_index.isin(summary_data.index)]
            summary_data = summary_data.loc[reorder_index]

        # Insert column of stat descriptions - we're going to leave this column name as a blank to avoid
        # having to prevent column name duplication (allow_duplicates is a newer pandas argument, and can't be sure it's available)
        summary_data.insert(0, '', summary_data.index)

        data = summary_data.to_records(index=False)
        columns = list(summary_data.columns)

    return data, columns

def process_column_list(mixed_list, M_c, column_lists, dedupe=False):
    output = []
    for identifier in mixed_list:
        if identifier == '*':
            output += get_all_column_names_in_original_order(M_c)
        elif column_lists != None and identifier in column_lists.keys():
            output += column_lists[identifier]
        else:
            output.append(identifier)
    if dedupe == True:
        check_for_duplicate_columns(output)
    return output



#
#   Copyright (c) 2010-2014, MIT Probabilistic Computing Project
#
#   Lead Developers: Jay Baxter and Dan Lovell
#   Authors: Jay Baxter, Dan Lovell, Baxter Eaves, Vikash Mansinghka
#   Research Leads: Vikash Mansinghka, Patrick Shafto
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import numpy
import os
import re
import inspect
import ast
import pylab
import matplotlib.cm
import time

import data_utils as du
import select_utils
import functions
import utils
import parser

def parse_pairwise_function(function_name, column=True, M_c=None, column_lists={}): ##TODO move to parser
    if column:
        if function_name == 'mutual information':
            return functions._mutual_information
        elif function_name == 'dependence probability':
            return functions._dependence_probability
        elif function_name == 'correlation':
            return functions._correlation
        else:
            raise utils.BayesDBParseError('Invalid column function: %s' % function_name)
    else:
        # TODO: need to refactor to support similarity with respect to column, because then we need to parse
        # and return the column id here.
        ##TODO temporary hack - move to parser maybe combine with parse_functions
        p = parser.Parser()
        _, target_columns = p.get_args_similarity(function_name, M_c, None, M_c, None, column_lists)
        if target_columns is None:
            return (functions._similarity, None)
        elif type(target_columns) == list:
            return (functions._similarity, target_columns)
        else:
            raise utils.BayesDBParseError('Invalid row function: %s' % function_name)

def get_columns(column_names, M_c): ##TODO move to parser or utils
    # If using a subset of the columns, get the appropriate names, and figure out their indices.
    if column_names is not None:
        column_indices = [M_c['name_to_idx'][name] for name in column_names]
    else:
        num_cols = len(M_c['name_to_idx'].keys())
        column_names = [M_c['idx_to_name'][str(idx)] for idx in range(num_cols)]
        column_indices = range(num_cols)
    column_names = numpy.array(column_names)
    return column_names, column_indices

def compute_raw_column_pairwise_matrix(function, X_L_list, X_D_list, M_c, T, engine, column_indices=None, numsamples=None):
    # Compute unordered matrix: evaluate the function for every pair of columns
    # Shortcut: assume all functions are symmetric between columns, only compute half.
    num_cols = len(column_indices)
    matrix = numpy.zeros((num_cols, num_cols))
    for i, orig_i in enumerate(column_indices):
        for j in range(i, num_cols):
            orig_j = column_indices[j]
            func_val = function((orig_i, orig_j), None, None, M_c, X_L_list, X_D_list, T, engine, numsamples)
            matrix[i][j] = func_val
            matrix[j][i] = func_val
    return matrix

def compute_raw_row_pairwise_matrix(function, arg, X_L_list, X_D_list, M_c, T, engine, row_indices=None, numsamples=None):
    # TODO: currently assume that the only function possible is similarity
    if row_indices is None:
        row_indices = range(len(T))
    num_rows = len(row_indices)
    matrix = numpy.zeros((num_rows, num_rows))
    for i, orig_i in enumerate(row_indices):
        for j in range(i, num_rows):
            orig_j = row_indices[j]
            func_val = function((orig_i, arg), orig_j, None, M_c, X_L_list, X_D_list, T, engine, numsamples)
            matrix[i][j] = func_val
            matrix[j][i] = func_val
    return matrix

def reorder_indices_by_cluster(matrix):
    # Hierarchically cluster columns.
    from scipy.spatial.distance import pdist
    from scipy.cluster.hierarchy import linkage, dendrogram
    Y = pdist(matrix)
    Z = linkage(Y)
    pylab.figure()
    dendrogram(Z)
    intify = lambda x: int(x.get_text())
    reorder_indices = map(intify, pylab.gca().get_xticklabels())
    pylab.clf() ## use instead of close to avoid error spam
    matrix_reordered = matrix[:, reorder_indices][reorder_indices, :]    
    return matrix_reordered, reorder_indices

def get_connected_clusters(matrix, cluster_threshold):
    # If cluster_threshold isn't none, then we want to return all the connected clusters
    # of columns: columns are connected if their edge weight is above the threshold.
    # Just do a search, starting at each column id.

    from collections import defaultdict
    clusters = [] # list of lists (conceptually a set of sets, but faster here)

    # Construct graph, in the form of a neighbor dictionary
    neighbors_dict = defaultdict(list)
    for i in range(matrix.shape[0]):
        for j in range(i+1, matrix.shape[0]):
            if matrix[i][j] > cluster_threshold:
                neighbors_dict[i].append(j)
                neighbors_dict[j].append(i)

    # Outer while loop: make sure every column has been visited
    unvisited = set(range(matrix.shape[0]))
    while(len(unvisited) > 0):
        cluster = []
        stack = [unvisited.pop()]
        while(len(stack) > 0):
            cur = stack.pop()
            cluster.append(cur)
            neighbors = neighbors_dict[cur]
            for n in neighbors:
                if n in unvisited:
                    stack.append(n)
                    unvisited.remove(n)                        
        if len(cluster) > 1:
            clusters.append(cluster)
    return clusters

def generate_pairwise_column_matrix(function_name, X_L_list, X_D_list, M_c, T, tablename='', limit=None, engine=None, column_names=None, cluster_threshold=None, numsamples=None):
    """
    Compute a matrix. In using a function that requires engine (currently only
    mutual information), engine must not be None.
    """

    # Get appropriate function
    function = parse_pairwise_function(function_name, column=True, M_c=M_c)

    # Get appropriate column information from column_names
    column_names, column_indices = get_columns(column_names, M_c)

    # Actually compute each function between each pair of columns
    matrix = compute_raw_column_pairwise_matrix(function, X_L_list, X_D_list, M_c, T, engine, column_indices, numsamples)

    if cluster_threshold is not None:
        # clusters is a list of lists, where the inner list contains the ids (into the matrix)
        # of the columns in each cluster.
        clusters = get_connected_clusters(matrix, cluster_threshold)
        
        # Now, convert the clusters from their matrix indices to their btable indices
        new_comps = []
        for comp in clusters:
            new_comps.append([column_indices[c] for c in comp])
        clusters = new_comps
    else:
        clusters = None

    # reorder the matrix
    matrix, reorder_indices = reorder_indices_by_cluster(matrix)
    column_names_reordered = column_names[reorder_indices]
            
    return matrix, column_names_reordered, clusters

def generate_pairwise_row_matrix(function_name, X_L_list, X_D_list, M_c, T, tablename='', engine=None, row_indices=None, cluster_threshold=None, column_lists={}, numsamples=None):
    """
    Compute a matrix. In using a function that requires engine (currently only
    mutual information), engine must not be None.
    """

    # Get appropriate function
    function, arg = parse_pairwise_function(function_name, column=False, M_c=M_c, column_lists=column_lists)

    # Get appropriate row list
    if row_indices is None:
        row_indices = numpy.array(range(len(T)))
    else:
        row_indices = numpy.array(row_indices)

    # Actually compute each function between each pair of columns
    matrix = compute_raw_row_pairwise_matrix(function, arg, X_L_list, X_D_list, M_c, T, engine, row_indices, numsamples)

    if cluster_threshold is not None:
        # clusters is a list of lists, where the inner list contains the ids (into the matrix)
        # of the columns in each cluster.
        clusters = get_connected_clusters(matrix, cluster_threshold)
        
        # Now, convert the clusters from their matrix indices to their btable indices
        new_comps = []
        for comp in clusters:
            new_comps.append([row_indices[c] for c in comp])
        clusters = new_comps
    else:
        clusters = None

    # reorder the matrix
    matrix, reorder_indices = reorder_indices_by_cluster(matrix)
    row_indices_reordered = row_indices[reorder_indices]
            
    return matrix, row_indices_reordered, clusters
    



#
#   Copyright (c) 2010-2014, MIT Probabilistic Computing Project
#
#   Lead Developers: Jay Baxter and Dan Lovell
#   Authors: Jay Baxter, Dan Lovell, Baxter Eaves, Vikash Mansinghka
#   Research Leads: Vikash Mansinghka, Patrick Shafto
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import time
import inspect
import sys
import pickle
import os
import numpy
import pytest
import random
import shutil
import pandas
from cStringIO import StringIO

import bayesdb.utils as utils
from bayesdb.client import Client
from bayesdb.engine import Engine
import bayesdb.bql_grammar as bql

test_tablenames = None
client = None
test_filenames = None

def setup_function(function):
  global test_tablenames, client, test_filenames
  test_tablenames = []
  test_filenames = []
  # Default upgrade_key_column is None, to let the user choose, but need to avoid
  # user input during testing, so for testing just create a new key column.
  client = Client(testing=True)

def teardown_function(function):
  global tablename, client
  for test_tablename in test_tablenames:
    client.engine.drop_btable(test_tablename)

  for test_filename in test_filenames:
    if os.path.exists(test_filename):
      os.remove(test_filename)

def update_cyclic_schema(tablename):
  global client
  for col in range(10):
    client( 'update schema for %s set col_%i=cyclic(0, 6.284)' % (tablename,col), debug=True, pretty=False )

def create_cyclic(path='data/cyclic_test.csv', key_column=0):
  test_tablename = 'cyclictest' + str(int(time.time() * 1000000)) + str(int(random.random()*10000000))
  csv_file_contents = open(path, 'r').read()
  client('create btable %s from %s' % (test_tablename, path), debug=True, pretty=False, key_column=key_column)

  global test_tablenames
  test_tablenames.append(test_tablename)

  return test_tablename

def test_cyclic_initalize():
  """test a few functions with cyclic data"""
  test_tablename = create_cyclic(key_column=0)
  global client, test_filenames

  # update schema
  update_cyclic_schema(test_tablename)

  # intialize model
  models = 3
  out = client('initialize %d models for %s' % (models, test_tablename), debug=True, pretty=False)[0]


def test_cyclic_all_the_things():
  """test a few functions with cyclic data"""
  test_tablename = create_cyclic(key_column=0)
  global client, test_filenames

  # update schema
  update_cyclic_schema(test_tablename)

  # intialize model
  models = 3
  out = client('initialize %d models for %s' % (models, test_tablename), debug=True, pretty=False)[0]

  # analyze
  iterations = 2
  client('analyze %s for %i iterations wait' % (test_tablename, iterations), debug=True, pretty=False)

  # select
  out = client('select col_0 from %s' % (test_tablename), pretty=False)[0]

  # predictive probability
  client('select predictive robability of col_0 from %s LIMIT 10' % (test_tablename), pretty=False)

  # simulate
  client('simulate col_0 from %s times 100' % (test_tablename), pretty=False)

  # simulate w/ GIVEN
  client('simulate col_0 GIVEN col_1 = 3.14 from %s times 100' % (test_tablename), pretty=False)
